/* CSV scanner, for tests */

%{
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "../unity.h"

#define TOKEN_NEWLINE 1
#define TOKEN_COMMA 2
#define TOKEN_STRING 3
static char* lval = "";

%}

%option noyywrap
%option nounput
%option prefix="testyy"

WHITESPACE	[ \t]
NEWLINECHAR	[\n\r]
NONCOMMA	[^,#\n\r]
%%

{NEWLINECHAR}			{ return TOKEN_NEWLINE; }
#.*				{ return TOKEN_NEWLINE; }
{WHITESPACE}*,{WHITESPACE}*	{ return TOKEN_COMMA; }
{NONCOMMA}+			{ lval = testyytext; return TOKEN_STRING; }
     
%%

// generous buffer size
#define BUFLEN 1024

void Usage(void);

const char* progname;

int main(int argc, char** argv) 
{
    int token;
    const char* filename;
    FILE* csv;
    char* input = NULL;
    UnitExpression *unit_expr = 0;
    char should_fail_p = 0;     /* should_fail_p true if this test-case should fail */
    char should_fail_msg[BUFLEN];
    int ngood = 0;
    int nbad = 0;
    char actual_parse_result[BUFLEN];
    char expected[BUFLEN];
    int column_number;
    const char* parser = "fits";
    char expected_formats_list[BUFLEN+1];
    char* expected_formats_p = NULL;
    int expected_formats_len = 0;
    int expected_status;
    int parser_id;

    progname = argv[0];
    {
        int ch;
        while ((ch = getopt(argc, argv, "p:")) >= 0) {
            switch (ch) {
              case 'p':
                parser = optarg; /* can I store this pointer? */
                break;
              default:
                Usage();
                break;
            }
        }
        argc -= optind;
        argv += optind;
    }
    if (argc != 1) {
        Usage();
    }

    filename = argv[0];

    if ((csv = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "%s: can't open file %s to read\n", progname, filename);
        exit(1);
    }

    testyyin = csv;
    column_number = 0;
    should_fail_p = 0;
    expected_formats_p = NULL;
    memset((void*)should_fail_msg, 0, BUFLEN);

    // expected_status indicates further tests:
    //   0: no further tests
    //   1: includes unrecognised units
    //   2: all units recognised, but at least one deprecated
    //   3: uses SI prefixes where forbidden
    //   4: all constraints satisfied
    expected_status = 0;
    
    parser_id = unity_identify_parser(parser);
    if (parser_id < 0) {
        fprintf(stderr, "Unidentified parser: %s\n", parser);
        exit(1);
    }

    while ((token = testyylex()) != 0) {

        switch (token) {
          case TOKEN_NEWLINE:
            if (column_number == 0) {
                // empty line
                break;
            }
            if (unit_expr != NULL) {
                unity_write_formatted(actual_parse_result, BUFLEN, unit_expr, "debug");
            }
            if (should_fail_p) {
                // this shouldn't have parsed
                if (unit_expr == NULL) {
                    ngood++;
                } else {
                    printf("Parsing: %s    Should have failed (%s), but produced: %s\n",
                           input, should_fail_msg, actual_parse_result);
                    nbad++;
                }
            } else {
                // this should have parsed
                if (unit_expr == NULL) {
                    // didn't parse
                    printf("Parsing: %s    Failed to parse\n", input);
                    nbad++;
                } else if (strcmp(expected, actual_parse_result) == 0) {
                    // this test passed
                    ngood++;
                    switch (expected_status) {
                      case 0:
                        break;  /* no other checks required */
                      case 1:
                        if (unity_check_expression(unit_expr, parser_id, UNITY_CHECK_RECOGNISED)) {
                            printf("Parsing: %s    has non-recognised units; didn't spot them!\n", input);
                            nbad++;
                        } else {
                            ngood++;
                        }
                        break;
                      case 2:
                        if (unity_check_expression(unit_expr, parser_id, UNITY_CHECK_RECOMMENDED)) {
                            printf("Parsing: %s    has deprecated units; didn't spot them!\n", input);
                            nbad++;
                        } else {
                            ngood++;
                        }
                        break;
                      case 3:
                        if (unity_check_expression(unit_expr, parser_id, UNITY_CHECK_CONSTRAINTS)) {
                            printf("Parsing: %s    has non-conformant units; didn't spot them!\n", input);
                            nbad++;
                        } else {
                            ngood++;
                        }
                        break;
                      case 4:
                        if (! unity_check_expression(unit_expr, parser_id, UNITY_CHECK_ALL)) {
                            printf("Parsing: %s    should have passed all tests, but didn't!\n", input);
                            nbad++;
                        } else {
                            ngood++;
                        }
                        break;
                      default:
                        fprintf(stderr, "Impossible expected_status=%d\n", expected_status);
                        exit(1);
                    }
                } else {
                    // parsed, but didn't produce expected output
                    printf("Parsing: %s    expected: %s\n    actual:   %s\n",
                           input, expected, actual_parse_result);
                    nbad++;
                }
            }
            if (unit_expr && expected_formats_p) {
                // check this parsed expression formats OK
                char actual_format[BUFLEN];
                char* syntax;
                for (syntax = strtok(expected_formats_list, ":");
                     syntax;
                     syntax = strtok(NULL, ":")) {
                    char* expected_format = strtok(NULL, ":");
                    if (expected_format == NULL) {
                        fprintf(stderr, "Malformed test file: expected formats unbalanced\n");
                        exit(1);
                    }
                    unity_write_formatted(actual_format, BUFLEN, unit_expr, syntax);
                    if (strcmp(actual_format, expected_format) == 0) {
                        ngood++;
                    } else {
                        printf("Formatting (%s): expected: %s\n    actual: %s\n",
                               syntax, expected_format, actual_format);
                        nbad++;
                    }
                }
            }
            // all done!

            if (unit_expr != NULL) {
                unity_free_expression(unit_expr);
            }

            // prepare for next time
            column_number = 0; // reset column number, so we start another check
            should_fail_p = 0; // reset to default false (if there's no second field)
            expected_formats_p = NULL;
            break;

          case TOKEN_COMMA:
            column_number++;
            break;
            
          case TOKEN_STRING: 
            switch (column_number) {
              case 0:
                // First item on CSV line
                input = lval;
                unit_expr = unity_parse_string(lval, parser_id);
                break;
              case 1:
                // Non-empty second column -- this input should fail to parse.
                // Copy the lval to the should_fail (explanation) buffer.
                should_fail_p = 1;
                strncpy(should_fail_msg, lval, BUFLEN-1);
                break;
              case 2:
                if (isdigit(lval[0]) || lval[0] == '-') {
                    // OK
                    expected_status = 0;
                    strncpy(expected, lval, BUFLEN-1);
                } else {
                    strncpy(expected, &lval[1], BUFLEN-2);
                    switch (lval[0]) {
                      case 'x':
                        expected_status = 1;
                        break;
                      case '?':
                        expected_status = 2;
                        break;
                      case 's':
                        expected_status = 3;
                        break;
                      case '!':
                        expected_status = 4;
                        break;
                      default:
                        fprintf(stderr, "Malformed test: unrecognised flag in %s", lval);
                        exit(1);
                    }
                }
                break;
              case 3:
                expected_formats_p = expected_formats_list;
                expected_formats_len = BUFLEN;
                strncpy(expected_formats_p, lval, expected_formats_len);
                expected_formats_p += strlen(lval);
                expected_formats_len -= strlen(lval);
                break;
              default:
                *expected_formats_p++ = ':';
                expected_formats_len--;
                strncpy(expected_formats_p, lval, expected_formats_len);
                expected_formats_p += strlen(lval);
                expected_formats_len -= strlen(lval);
                break;
            }
            break;

          default:
            fprintf(stderr, "Impossible token: %d\n", token);
            break;
        }
    }

    if (nbad == 0) {
        printf("%6s : ALL %d tests pass\n", parser, ngood);
    } else {
        printf("%6s : Passes: %d\n         FAILS:  %d\n", parser, ngood, nbad);
    }

    return 0;
}

void Usage(void) 
{
    fprintf(stderr, "Usage: %s [-p parser] csv-file\n", progname);
    exit(1);
}

