/**
 * @file unity.h
 */

#include <stdlib.h>
#include <string.h>

#include "unit-definitions.h"

// defines:
//   UnitDef unit_defs[];
//   int unit_defs_length;
//   #define UNITY_SYNTAX_*
#include "known-units-cinclude.c"

static int unit_def_compare(const void* key, const void* test) 
{
    return strcmp((char*)key, ((UnitDef*)test)->abbrev);
}

/**
 * Retrieve a unit definition
 * @param abbrev a unit abbreviation, for example "m" for "metre"
 * @return a pointer to a the unit definition which this abbreviation corresponds to
 */
const UnitDef* unity_get_unit(const char* abbrev) 
{
    return (UnitDef*)bsearch((void*)abbrev,
                             (void*)&unit_defs[0], unit_defs_length, sizeof(UnitDef),
                             &unit_def_compare);
}

/**
 * Indicates whether the unit is being used in a way which satisfies
 * the indicated constraints.  The checks are indicated by the bitwise
 * OR of the following flags:
 *
 * <p><code>UNITY_CHECK_RECOGNISED</code>: the unit is a recognised unit, in the
 * sense that it is listed in the specification of the corresponding
 * syntax.
 *
 * <p><code>UNITY_CHECK_RECOMMENDED</code>: the unit is a recognised unit, and is
 * additionally not listed in the specification as deprecated.
 *
 * <p><code>UNITY_CHECK_CONSTRAINTS</code>: the unit is being used in conformance
 * with any other constraints placed on it.  Most typically, it either
 * has no SI prefix, or is allowed to have such a prefix.
 *
 * <p>The constant <code>UNITY_CHECK_ALL</code> performs all checks.
 *
 * @param u the Unit to be checked
 * @param syntax one of the constants <code>UNITY_SYNTAX_FITS</code>
 * or <code>..._OGIP</code> or <code>..._CDS</code>
 * @param flags the checks to be performed
 * @return true (non-zero) if the checks pass
 */
int unity_check_unit(const Unit* u, const int syntax, const int flags) 
{
    const UnitDef* def;
    int is_recognised;
    int is_recommended;
    int satisfies_constraints;
    int result;

    if (flags == 0) {
        // trivial check
        return 1;               /* JUMP OUT */
    }

    def = u->base_unit_def;
    if (def == NULL) {
        is_recognised = 0;
        is_recommended = 0;
        satisfies_constraints = 1; /* no constraints, so all satisfied */
    } else {
        const UnitDefSyntax* s;
        if (syntax < 0 || syntax >= U_NSYNTAXES) {
            fprintf(stderr, "unity_check_unit: Illegal syntax specifier\n");
            return 0;           /* JUMP OUT */
        }
        s = &(def->syntaxes[syntax]);
        if (s->syntax_name == NULL) {
            // this syntax does not recognise this unit
            is_recognised = 0;
            is_recommended = 0;
            satisfies_constraints = 1; /* no constraints, so all satisfied */
        } else {
            is_recognised = 1;
            is_recommended = !s->is_deprecated;
            satisfies_constraints = u->prefix_power == 0 || s->may_have_si_prefixes;
        }
    }

    result = 1;                 /* true */
    if (flags & UNITY_CHECK_RECOGNISED)
        result &= is_recognised;
    if (flags & UNITY_CHECK_RECOMMENDED)
        result &= is_recommended;
    if (flags & UNITY_CHECK_CONSTRAINTS)
        result &= satisfies_constraints;
    return result;
}

/**
 * Indicates whether the units in the expression are being used in a way which satisfies
 * the indicated constraints.  This checks each unit using 
 * {@link #unity_check_unit}.
 *
 * @param ue the UnitExpression to be checked
 * @param syntax one of the constants <code>UNITY_SYNTAX_FITS</code>
 * or <code>..._OGIP</code> or <code>..._CDS</code>
 * @param flags the checks to be performed
 * @return true (non-zero) if the checks pass
 */
int unity_check_expression(const UnitExpression* ue, const int syntax, const int flags) 
{
    Unit* u;
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        if (! unity_check_unit(u, syntax, flags)) {
            return 0;           /* false */
        }
    }
    return 1;
}
    
#if 0
    /**
     * Indicates whether the base unit is one of those recommended
     * within the specification of the given syntax.  In this context,
     * "recommended" means "mentioned in the specification", so deprecated units
     * count as recommended ones.
     *
     * <p>Note that this checks that the <em>unit</em> is a
     * recommended one: we don't (currently) check whether the
     * abbreviation that got us here is a recommended one (for
     * example, "pixel" is a valid FITS/CDS name for pixels, and "pix"
     * is a FITS and OGIP one).
     * @param ue the unit-expression to be checked
     * @param syntax one of the syntaxes of <code>*_SYNTAX</code>
     * @return true (1) if all of the units in the expression are recommended ones
     * @see #unity_has_deprecated
     */
int unity_all_recommended(const UnitExpression* ue, const char* syntax) 
{
    Unit* u;
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        int i;
        int unit_is_ok = 0;
        const UnitDef* def = u->base_unit_def;
        if (def == NULL) {
            return 0;           /* JUMP OUT */
        }
        for (i=0; i<U_NSYNTAXES; i++) {
            if (strcmp(def->syntaxes[i].syntax_name, syntax) == 0) {
                // that's fine -- this is a recommended unit in this syntax
                unit_is_ok = 1;
                break;          /* break one level, to next Unit */
            }
        }
        if (! unit_is_ok) {
            // we haven't discovered that this unit is OK in this syntax,
            // so the function evaluates to false overall
            return 0;           /* JUMP OUT */
        }
    }
    // we haven't broken out already, so everything's OK
    return 1;
}

/**
 * Indicates whether the parsed expression has any deprecated
 * units, where a "deprecated" unit is a known unit specifically
 * marked as deprecated.  That is, an expression composed of only
 * non-deprecated and unknown units has no deprecated units.
 * @param ue the unit-expression to be checked
 * @param syntax the syntax with respect to which the
 * units should be checked
 * @see #unity_all_recommended
 * @return true (1) if there is at least one unit marked as deprecated
 */
int unity_has_deprecated(const UnitExpression* ue, const char* syntax) 
{
    Unit* u;

    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        int i;
        const UnitDef* def = u->base_unit_def;
        if (def == NULL) {
            // not deprecated
            continue;
        }
        for (i=0; i<U_NSYNTAXES; i++) {
            if (strcmp(def->syntaxes[i].syntax_name, syntax) == 0) {
                // this is a recommended unit in this syntax
                if (def->syntaxes[i].is_deprecated) {
                    // that's all we need to know...
                    return 1;   /* JUMP OUT */
                }
            }
        }
    }
    // if we get to the end, then we didn't find any deprecated units
    return 0;
}
#endif
