#ifndef UNITY_H_LOADED
#define UNITY_H_LOADED 1
/* unitylib.h ... */
#ifndef UNITYLIB_H_LOADED
#define UNITYLIB_H_LOADED 1

/**
 * @file unity.h
 */

/* forward definition -- see unit-definition.h */
struct unit_definition;

/** @struct unit_struct
 * A component of a unit expression
 */
struct unit_struct {
    /**
     * The base-10 logarithm of the prefix in front of the unit.  For
     * example, this will be '-3' for the 'm' (milli-) prefix.
     */
    int prefix_power;           /* -24 to +24 */
    /**
     * The base unit of this unit specification, as a known unit.
     */
    const struct unit_definition* base_unit_def;
    /**
     * The base unit of this unit specification.  This might be 'm'
     * for metres, 'Hz' for Hertz, and so on.
     */
    char* base_unit_string;     /* eg Hz or aangstrom */
    /**
     * The power to which the unit is raised.  The specification
     * 'mm^(-2)' will result in a power of '-2'.
     */
    float power;
    /** A pointer to the next unit in the sequence */
    struct unit_struct *next;
};
/** A clearer name for struct unit_struct */
typedef struct unit_struct Unit;

/** @struct UnitExpression
 * The parsed version of a unit string.
 */
typedef struct {
    /**
     * The base-10 log of the numerical factor at the head of the string. For some
     * parsers, this is restricted to round powers of ten.
     */
    float log_factor;
    /**
     * The sequence of units which comprise the parsed units specification.
     */
    Unit* unit_sequence;
} UnitExpression;

UnitExpression* unity_parse_string(const char* unit_string, const int syntax);
/* unity_parse_error is defined in lex.lex */
/**
 * Retrieve the most recent lexing or parsing error.  Immediately after
 * the parser has returned with an error (that is function
 * 'unity_parse_string' has returned NULL) this function can be called
 * to obtain an explanation.
 * @sa unity_parse_string
 * @return a static string containing an explanation of a parse error
 */
const char* unity_parse_error(void);
const char** unity_parser_names(void);

/** Indicates the 'FITS' syntax */
#define UNITY_SYNTAX_FITS 0
/** Indicates the 'OGIP' syntax */
#define UNITY_SYNTAX_OGIP 1
/** Indicates the 'CDS' syntax */
#define UNITY_SYNTAX_CDS 2

/* The number of syntaxes */
#define U_NSYNTAXES 3

int unity_identify_parser(const char* parser_name);
const char* unity_parser_name(const int parser_id);
void unity_free_expression(UnitExpression* u);
const char* unity_version_string();
int unity_version_number();

#ifdef UNITY_INTERNAL
/* Internal functions */
Unit* u_new_unit(const char* unitString, const float power);
Unit* u_unit_append(Unit* u1, Unit* u2);
void u_receive_result(float, Unit* unit_sequence);
int u_prefix_to_power(const char pfx);
//char* u_copy_string(const char c, const char* s);
Unit* u_divide_units(Unit* num, Unit* den);
Unit* u_unit_reciprocal(Unit* u);
void u_free_unit(Unit* u);
#endif

#endif
/* ...end unitylib.h */
/* writers.h ... */
#ifndef WRITERS_H_LOADED
#define WRITERS_H_LOADED 1

#include <stdio.h>

#ifdef UNITY_INTERNAL
typedef char* (*writef_ptr) (char*, int, const UnitExpression*);
#endif

char* unity_write_formatted(char* buf, int buflen, const UnitExpression*, const char*);
const char** unity_formatter_names(void);

#endif
/* ...end writers.h */
/* unit-definitions.h ... */
#ifndef UNIT_DEFINITIONS_H
#define UNIT_DEFINITIONS_H 1


struct unit_definition_syntax {
    const char* syntax_name;    /* the name of the syntax being described (eg "fits") */
    char may_have_si_prefixes;
    char is_deprecated;
    const char* preferred_abbrev; /* such as 'm' for metre */
};
typedef struct unit_definition_syntax UnitDefSyntax;

struct unit_definition {
    const char* abbrev; /* the name of one of the abbrevs for this unit */
    const char* name;   /* the name of the unit, such as 'metre' -- not the abbreviation 'm' */
    const char* type;   /* type such as 'length' */
    const char* notes;  /* for example 'Hz = s-1', NULL if no notes present */
    const char* latex_form;     /* an abbrev suitable for LaTeX, if different from a preferred syntax */
    /* There is a unit_definition element for each syntax, whether or
       not the unit is recognised.  If a syntax does not recognise a
       unit, this is indicated by setting syntax_name to NULL. */
    struct unit_definition_syntax syntaxes[U_NSYNTAXES];
};
typedef struct unit_definition UnitDef;

const UnitDef* unity_get_unit(const char* abbrev);

#define UNITY_CHECK_RECOGNISED 1
#define UNITY_CHECK_RECOMMENDED 2
#define UNITY_CHECK_CONSTRAINTS 4
#define UNITY_CHECK_ALL UNITY_CHECK_RECOGNISED | UNITY_CHECK_RECOMMENDED | UNITY_CHECK_CONSTRAINTS
int unity_check_unit(const Unit* u, const int syntax, const int flags);

int unity_check_expression(const UnitExpression* ue, const int syntax, const int flags);

#endif /* UNIT_DEFINITIONS_H */
/* ...end unit-definitions.h */
#endif
