#ifndef UNITYLIB_H_LOADED
#define UNITYLIB_H_LOADED 1

/**
 * @file unity.h
 */

/* forward definition -- see unit-definition.h */
struct unit_definition;

/** @struct unit_struct
 * A component of a unit expression
 */
struct unit_struct {
    /**
     * The base-10 logarithm of the prefix in front of the unit.  For
     * example, this will be '-3' for the 'm' (milli-) prefix.
     */
    int prefix_power;           /* -24 to +24 */
    /**
     * The base unit of this unit specification, as a known unit.
     */
    const struct unit_definition* base_unit_def;
    /**
     * The base unit of this unit specification.  This might be 'm'
     * for metres, 'Hz' for Hertz, and so on.
     */
    char* base_unit_string;     /* eg Hz or aangstrom */
    /**
     * The power to which the unit is raised.  The specification
     * 'mm^(-2)' will result in a power of '-2'.
     */
    float power;
    /** A pointer to the next unit in the sequence */
    struct unit_struct *next;
};
/** A clearer name for struct unit_struct */
typedef struct unit_struct Unit;

/** @struct UnitExpression
 * The parsed version of a unit string.
 */
typedef struct {
    /**
     * The base-10 log of the numerical factor at the head of the string. For some
     * parsers, this is restricted to round powers of ten.
     */
    float log_factor;
    /**
     * The sequence of units which comprise the parsed units specification.
     */
    Unit* unit_sequence;
} UnitExpression;

UnitExpression* unity_parse_string(const char* unit_string, const int syntax);
/* unity_parse_error is defined in lex.lex */
/**
 * Retrieve the most recent lexing or parsing error.  Immediately after
 * the parser has returned with an error (that is function
 * 'unity_parse_string' has returned NULL) this function can be called
 * to obtain an explanation.
 * @sa unity_parse_string
 * @return a static string containing an explanation of a parse error
 */
const char* unity_parse_error(void);
const char** unity_parser_names(void);

/** Indicates the 'FITS' syntax */
#define UNITY_SYNTAX_FITS 0
/** Indicates the 'OGIP' syntax */
#define UNITY_SYNTAX_OGIP 1
/** Indicates the 'CDS' syntax */
#define UNITY_SYNTAX_CDS 2

/* The number of syntaxes */
#define U_NSYNTAXES 3

int unity_identify_parser(const char* parser_name);
const char* unity_parser_name(const int parser_id);
void unity_free_expression(UnitExpression* u);
const char* unity_version_string();
int unity_version_number();

#ifdef UNITY_INTERNAL
/* Internal functions */
Unit* u_new_unit(const char* unitString, const float power);
Unit* u_unit_append(Unit* u1, Unit* u2);
void u_receive_result(float, Unit* unit_sequence);
int u_prefix_to_power(const char pfx);
//char* u_copy_string(const char c, const char* s);
Unit* u_divide_units(Unit* num, Unit* den);
Unit* u_unit_reciprocal(Unit* u);
void u_free_unit(Unit* u);
#endif

#endif
