#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <assert.h>

/**
 * @file unity.h
 */

#define UNITY_INTERNAL 1

#include "writers.h"

#include "unit-definitions.h"

/* Support writing formatted text to a buffer.
 * Not thread safe.
 */
static int buf_nwritten = 0;
static int buf_buflen;
static char* buf_buf;
void bufinit(char* buf, int buflen) 
{
    buf_buf = buf;
    buf_buflen = buflen;
    buf_nwritten = 0;
    return;
}
char* bufprintf(char* fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    buf_nwritten += vsnprintf(&buf_buf[buf_nwritten], buf_buflen-buf_nwritten, fmt, ap);
    va_end(ap);

    return buf_buf;
}    

static int write_number(float f, int brackets_float, int brackets_negative)
{
    if (f == floorf(f)) {        /* no fractional part */
        int i = (int)roundf(f);
        bufprintf((brackets_negative && i<0 ? "(%d)" : "%d"), i);
    } else {
        int brackets = brackets_float || (f < 0.0 && brackets_negative);
        bufprintf((brackets ? "(%g)" : "%g"), f);
    }
    return 0;
}

static const char* prefix_power_to_string(int prefixPower) {
    const char* rval;
    switch (prefixPower) {
      case -24: rval = "y"; break;
      case -21: rval = "z"; break;
      case -18: rval = "a"; break;
      case -15: rval = "f"; break;
      case -12: rval = "p"; break;
      case  -9: rval = "n"; break;
      case  -6: rval = "u"; break;
      case  -3: rval = "m"; break;
      case  -2: rval = "c"; break;
      case  -1: rval = "d"; break;
      case   0: rval = "";  break;
      case   1: rval = "da";break;
      case   2: rval = "h"; break;
      case   3: rval = "k"; break;
      case   6: rval = "M"; break;
      case   9: rval = "G"; break;
      case  12: rval = "T"; break;
      case  15: rval = "P"; break;
      case  18: rval = "E"; break;
      case  21: rval = "Z"; break;
      case  24: rval = "Y"; break;
      default:
        // This should be impossible, since it's only
        // prefixLetterToPower which sets these values
        assert(0);
    }
    return rval;
}

static const char* prefix_power_to_string_latex(int prefixPower) {
    const char* rval;
    switch (prefixPower) {
      case -24: rval = "\\yocto "; break;
      case -21: rval = "\\zepto "; break;
      case -18: rval = "\\atto ";  break;
      case -15: rval = "\\femto "; break;
      case -12: rval = "\\pico ";  break;
      case  -9: rval = "\\nano ";  break;
      case  -6: rval = "\\micro "; break;
      case  -3: rval = "\\milli "; break;
      case  -2: rval = "\\centi "; break;
      case  -1: rval = "\\deci ";  break;
      case   0: rval = "";         break;
      case   1: rval = "\\deca ";  break;
      case   2: rval = "\\hecto "; break;
      case   3: rval = "\\kilo ";  break;
      case   6: rval = "\\mega ";  break;
      case   9: rval = "\\giga ";  break;
      case  12: rval = "\\tera ";  break;
      case  15: rval = "\\peta ";  break;
      case  18: rval = "\\exa ";   break;
      case  21: rval = "\\zetta "; break;
      case  24: rval = "\\yotta "; break;
      default:
        // This should be impossible, since it's only
        // prefixLetterToPower which sets these values
        assert(0);
    }
    return rval;
}

static const char* find_base_unit(Unit* u, const int syntax) 
{
    if (u->base_unit_string != NULL) {
        return u->base_unit_string;
    } else {
        const char* unit_string = NULL;
        const UnitDef* def = u->base_unit_def;
        int i;
        assert(0 <= syntax && syntax < U_NSYNTAXES);
        for (i=0; i<U_NSYNTAXES; i++) {
            const char* this_syntax = def->syntaxes[i].syntax_name;
            if (this_syntax != NULL) {
                if (syntax == i) {
                    // this is the syntax we're looking for
                    unit_string = def->syntaxes[i].preferred_abbrev;
                    break;
                }
                if (unit_string == NULL) {
                    // make sure unit_string has something in it (acts as default if no match)
                    unit_string = def->syntaxes[i].preferred_abbrev;
                }
            }
        }
        assert(unit_string != NULL);
        return unit_string;
    }
}

static const char* find_base_unit_latex(Unit* u)
{
    const char* unit_string;
    if (u->base_unit_string != NULL) {
        unit_string = u->base_unit_string;
    } else {
        const UnitDef* def = u->base_unit_def;
        assert(def != NULL);
        if (def->latex_form == NULL) {
            unit_string = find_base_unit(u, UNITY_SYNTAX_FITS);
        } else {
            unit_string = def->latex_form;
        }
    }
    assert(unit_string != NULL);
    return unit_string;
}


char* write_fits(char* buf, int buflen, const UnitExpression* ue)
{
    int first = 1;
    Unit* u;

    bufinit(buf, buflen);
    
    if (ue->log_factor != 0.0) {
        bufprintf("10^");
        write_number(ue->log_factor, 1, 0);
        first = 0;
    }
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        if (first) {
            first = 0;
        } else {
            bufprintf(" ");
        }
        bufprintf("%s%s", prefix_power_to_string(u->prefix_power), find_base_unit(u, UNITY_SYNTAX_FITS));
        if (u->power != 1.0) {
            write_number(u->power, 1, 0);
        }
    }
    return buf;                   /* success return */
}

char* write_ogip(char* buf, int buflen, const UnitExpression* ue)
{
    int first = 1;
    Unit* u;

    bufinit(buf, buflen);

    if (ue->log_factor != 0.0) {
        bufprintf("10**");
        write_number(ue->log_factor, 0, 1);
        first = 0;
    }
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        const char* base_unit_string = find_base_unit(u, UNITY_SYNTAX_OGIP);
        float e = u->power;
        if (first) {
            first = 0;
        } else {
            bufprintf(" ");
        }
        if (e == 1.0) {
            bufprintf("%s%s", prefix_power_to_string(u->prefix_power), base_unit_string);
        } else if (e < 0.0) {
            bufprintf("/%s%s", prefix_power_to_string(u->prefix_power), base_unit_string);
            if (e != -1.0) {
                bufprintf("**");
                write_number(-e, 0, 1);
            }
        } else {
            // e is positive and not 1.0
            bufprintf("%s%s**", prefix_power_to_string(u->prefix_power), base_unit_string);
            write_number(e, 0, 1);
        }
    }
    return buf;                   /* success return */
}

char* write_cds(char* buf, int buflen, const UnitExpression* ue)
{
    int first = 1;
    Unit* u;

    bufinit(buf, buflen);

    if (ue->log_factor != 0.0) {
        float lfactor = ue->log_factor;
        if (lfactor == floor(lfactor) && (lfactor < -2.0 || lfactor > 2.0)) {
            bufprintf("10%s", (lfactor > 0 ? "+" : ""));
            write_number(lfactor, 0, 0);
        } else {
            bufprintf("%g", pow(10,lfactor));
        }
    }
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        const char* base_unit_string = find_base_unit(u, UNITY_SYNTAX_CDS);
        float e = u->power;
        if (e == 1.0) {
            bufprintf("%s%s%s", (first ? "" : "."), prefix_power_to_string(u->prefix_power), base_unit_string);
        } else if (e < 0.0) {
            bufprintf("/%s%s", prefix_power_to_string(u->prefix_power), base_unit_string);
            if (e != -1.0) {
                write_number(-e, 0, 0);
            }
        } else {
            // e is positive and not 1.0
            bufprintf("%s%s%s", (first ? "" : "."), prefix_power_to_string(u->prefix_power), base_unit_string);
            write_number(e, 0, 0);
        }
        first = 0;
    }
    return buf;                   /* success return */
}

char* write_latex(char* buf, int buflen, const UnitExpression* ue)
{
    int first = 1;
    Unit* u;

    bufinit(buf, buflen);

    if (ue->log_factor == 0.0) {
        bufprintf("\\si{");
    } else {
        bufprintf("\\SI{1e");
        write_number(ue->log_factor, 0, 0);
        bufprintf("}{");
    }
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        const char* base_unit_string = find_base_unit_latex(u);
        float e = u->power;
        if (e == 1.0) {
            bufprintf("%s%s%s", (first ? "" : "."), prefix_power_to_string_latex(u->prefix_power), base_unit_string);
        } else if (e < 0.0) {
            bufprintf("\\per %s%s", prefix_power_to_string_latex(u->prefix_power), base_unit_string);
            if (e != -1.0) {
                bufprintf("^{");
                write_number(-e, 0, 0);
                bufprintf("}");
            }
        } else {
            // e is positive and not 1.0
            bufprintf("%s%s%s", (first ? "" : "."), prefix_power_to_string_latex(u->prefix_power), base_unit_string);
            bufprintf("^{");
            write_number(e, 0, 0);
            bufprintf("}");
        }
        first = 0;
    }
    bufprintf("}");
    return buf;                   /* success return */
}

char* write_debugging(char* buf, int buflen, const UnitExpression* unit_expression) 
{
     Unit* u;

     bufinit(buf, buflen);
     
    if (unit_expression->log_factor == 1.0) {
        bufprintf("1");
    } else {
        bufprintf("%.3g", unit_expression->log_factor);
    }
    for (u=unit_expression->unit_sequence; u!=NULL; u=u->next) {
        bufprintf(" %d/%s/%g", u->prefix_power, find_base_unit(u, UNITY_SYNTAX_FITS), u->power);
    }
    return buf;
}

static struct {
    const char* name;
    writef_ptr func;
} writer_list[] = {
    { "fits",  &write_fits },
    { "ogip",  &write_ogip },
    { "cds",   &write_cds },
    { "latex", &write_latex },
    { "debug", &write_debugging },
};
static int writer_list_size = sizeof(writer_list)/sizeof(writer_list[0]);

/**
 * Write a unit-expression to the given buffer.  The available formats
 * are those returned by the function {@link unity_parser_names}.
 *
 * @param buf the buffer to be written to
 * @param buflen the length of the input buffer
 * @param ue the unit-expression to be written
 * @param format one of the available formats, as a string
 * @return buf on success, or NULL on error
 */
char* unity_write_formatted(char* buf, int buflen, const UnitExpression* ue, const char* format)
{
    writef_ptr f = NULL;
    int i;
    char* rval;

    for (i=0; i<writer_list_size; i++) {
        if (strcmp(format, writer_list[i].name) == 0) {
            f = writer_list[i].func;
            break;
        }
    }
    if (f == NULL) {
        fprintf(stderr, "Can't find a writer of format %s\n", format);
        rval = NULL;
    } else {
        f(buf, buflen, ue);
        rval = buf;
    }
    return rval;
}

/**
 * Return the available formatter names.  If this is impossible for some
 * reason (which can really only be memory exhaustion, which should
 * never happen), print an error message and return NULL.
 * The returned list is statically allocated, and should not be freed
 * by the caller.
 * @return a NULL-terminated list of pointers to formatter names
 */
const char** unity_formatter_names(void)
{
    static const char** res = NULL;
    if (res == NULL) {
        // initialise
        int i;
        if ((res = (const char**)malloc(sizeof(char*)*(writer_list_size+1))) == NULL) {
            fprintf(stderr, "Can't allocate space for unity_format_names\n");
            return NULL;
        }
        for (i=0; i<writer_list_size; i++) {
            res[i] = writer_list[i].name;
        }
        res[writer_list_size] = NULL; /* terminate the list */
    }
    return res;
}
