/** 
 * Parse the known-units spreadsheet, and generate C/Java code
 * representing it.
 *
 * At present, the program only reads in the CSV, and dumps it in a
 * debug form.
 */

import uk.me.nxg.unity.UnitDefinition;
import uk.me.nxg.unity.UnitDefinitionMap;

//import au.com.bytecode.opencsv.CSVReader;
import java.io.FileReader;
import java.util.List;
import java.io.File;

import uk.me.nxg.unity.UnitParser;


public class ParseUnits {

    static private final String[] knownSyntaxes = { UnitParser.FITS, UnitParser.OGIP, UnitParser.CDS };
    
    public static void main(String[] args) {

        File csvFile = null;
        File javaOutputFile = null;
        File cOutputFile = null;
        char optarg = '\0';
        boolean verbose = false;

        for (String arg : args) {
            if (arg.charAt(0) == '-') {
                switch (arg.charAt(1)) {
                  case 'j':
                    optarg = 'j';
                    break;
                  case 'c':
                    optarg = 'c';
                    break;
                  case 'v':
                    verbose = true;
                    break;
                  default:
                    Usage();
                }
            } else {
                switch (optarg) {
                  case '\0':
                    if (csvFile != null) {
                        Usage();
                    } else {
                        csvFile = new File(arg);
                    }
                    break;
                  case 'j':
                    if (javaOutputFile != null) {
                        Usage();
                    } else {
                        javaOutputFile = new File(arg);
                    }
                    optarg = '\0';
                    break;
                  case 'c':
                    if (cOutputFile != null) {
                        Usage();
                    } else {
                        cOutputFile = new File(arg);
                    }
                    optarg = '\0';
                    break;
                  default:
                    Usage();
                }
            }
        }

        if (csvFile == null) {
            Usage();
        }
        assert csvFile != null;

        java.util.Map<String,UnitDefinition> defs = null;

        try {
            defs = readUnitdefs(csvFile);
        } catch (java.io.FileNotFoundException e) {
            System.err.println("Can't find file " + csvFile + " to read (" + e + ")");
            System.exit(1);
        } catch (Exception e) {
            System.err.println("Other exception: " + e);
            System.exit(1);
        }

        assert defs != null;

        if (javaOutputFile != null) {
            try {
                UnitDefinitionMap.persist(defs, javaOutputFile);
            } catch (java.io.IOException e) {
                System.err.println("IO Exception writing object-serialised file: " + e);
            }
        }

        if (cOutputFile != null) {
            try {
                dumpCDefinitions(defs, cOutputFile);
            } catch (java.io.IOException e) {
                System.err.println("Unable to write C definition file: " + e);
            }
        }   

        if (verbose) {
            for (String u : new java.util.TreeSet<String>(defs.keySet())) {
                UnitDefinition d = defs.get(u);
                System.out.print("Unit: " + u + " -> " + d.name() + ": ");
                for (String stxName : knownSyntaxes) {
                    UnitDefinition.Syntax stx = d.getSyntax(stxName);
                    if (stx == null) {
                        System.out.print("NO; ");
                    } else {
                        System.out.print(stx);
                    }
                }
                System.out.println();
            }
        }
        
    }

    static java.util.Map<String,UnitDefinition> readUnitdefs(File inputFile) 
            throws java.io.FileNotFoundException, uk.me.nxg.unity.UnitParserException {

        // an empty string contains only horizontal whitespace
        java.util.regex.Pattern emptyString = java.util.regex.Pattern.compile("^[ \\t]*$");

        //CSVReader reader = new CSVReader(new FileReader(inputFile));
        SimpleCSVReader reader = new SimpleCSVReader(new FileReader(inputFile));
        java.util.Map<String,UnitDefinition> defs = new java.util.HashMap<String,UnitDefinition>();

        UnitDefinition def = null;
        for (java.util.List<String> line : reader) {
            if (line.get(0).startsWith("#")) {
                // Skip comment lines
                continue;
            }
            //System.out.println("Line: " + line.get(0) + ":" + line.get(1));

            if (line.size() < 8) {
                System.err.println("Malformed CSV file: expected at least 8 columns, got " + line.size() + ":");
                for (int i=0; i<line.size(); i++) {
                    System.err.print('[' + line.get(i) + ']');
                }
                System.err.println();
                System.exit(1);
            }

            // If element line.get(1) is empty, then this line is
            // describing a new abbreviated representation of the
            // same unit (eg "pix" after "pixel" for units of
            // pixels).
            java.util.regex.Matcher m = emptyString.matcher(line.get(1));
            if (!m.matches()) {
                // This line.get(1) is not empty, so it's a
                // new unit definition.
                def = new UnitDefinition(line.get(1), line.get(2), line.get(3), line.get(4));
            }
            if (def == null) {
                System.err.println("Malformed CSV file: column 2 may not be empty on the first row");
                System.exit(1);
            }

            // associate this (new or current) definition with this abbreviation
            defs.put(line.get(0), def);
            // ... and 
            def.addSyntax("fits", line.get(0), line.get(5));
            def.addSyntax("ogip", line.get(0), line.get(6));
            def.addSyntax("cds",  line.get(0), line.get(7));
        }
        return defs;
    }

    static private void dumpCDefinitions(java.util.Map<String,UnitDefinition> defs, File cOutputFile)
            throws java.io.FileNotFoundException {
        java.io.PrintWriter out = new java.io.PrintWriter(cOutputFile);

        // The C structs being initialised are:
        //
        // struct unit_definition_syntax {
        //     const char* syntax_name;    /* the name of the syntax being described (eg "fits") */
        //     char may_have_si_prefixes;
        //     char is_deprecated;
        //     const char* preferred_abbrev; /* such as 'm' for metre */
        // };
        // typedef struct unit_definition_syntax UnitDefSyntax;
        //
        // struct unit_definition {
        //     const char* abbrev; /* the name of one of the abbrevs for this unit */
        //     const char* name;   /* the name of the unit, such as 'metre' -- not the abbreviation 'm' */
        //     const char* type;   /* type such as 'length' */
        //     const char* notes;  /* for example 'Hz = s-1', NULL if no notes present */
        //     const char* latex_form;     /* an abbrev suitable for LaTeX, if different from a preferred syntax */
        //     struct unit_definition_syntax syntaxes[U_NSYNTAXES];
        // };
        // typedef struct unit_definition UnitDef;
        //
        // The constants
        //   #define FITS_SYNTAX 0
        //   #define OGIP_SYNTAX 1
        //   #define CDS_SYNTAX 2
        // ...are initialised in unitylib.h.  The number of syntaxes,
        // and the indexes of the syntaxes within the
        // unit_definition.syntaxes array, as produced by this program, MUST match these indexes.

        out.println("#include <stdio.h>"); // for NULL
        out.printf("#if U_NSYNTAXES != %d\n#  error \"Mismatch in U_NSYNTAXES between ParseUnits and header\"\n#endif\n",
                   knownSyntaxes.length);
        out.println("static struct unit_definition unit_defs[] = {");

        // we write out an array of unit_definition elements, one for
        // each possible abbrev (that is those units which have two
        // abbrevs, either in multiple syntaxes or in one, are
        // effectively duplicated in this list).
        for (String abbrev : new java.util.TreeSet<String>(defs.keySet())) {
            UnitDefinition u = defs.get(abbrev);
            out.printf("    { \"%s\", \"%s\", \"%s\"", abbrev, u.name(), u.type());
            if (u.notes().length() == 0) {
                out.printf(", NULL");
            } else {
                out.printf(", \"%s\"", u.notes());
            }
            if (u.latexForm() == null) {
                out.printf(", NULL");
            } else {
                // escape backslashes in LaTeX forms
                out.printf(", \"%s\"", u.latexForm().replace("\\", "\\\\"));
            }
            out.printf(", {");
            for (String stxName : knownSyntaxes) {
                UnitDefinition.Syntax stx = u.getSyntax(stxName);
                if (stx == null) {
                    out.printf("\n        { NULL, 0, 0, NULL },");
                } else {
                    out.printf("\n        { \"%s\", %d, %d, \"%s\" },", stxName,
                               (stx.mayHaveSIPrefixes() ? 1 : 0),
                               (stx.isDeprecated() ? 1 : 0),
                               stx.getOneAbbrev()); // FIXME: guarantee preferred one?
                }
            }
            out.println("} },");
        }

        out.println("};");
        out.println("static int unit_defs_length = sizeof(unit_defs)/sizeof(unit_defs[0]);");

        out.close();
    }

    static private void Usage() {
        System.err.println("Usage: ParseUnits [-j outputfile] [-c outputfile] csv-filename");
        System.exit(1);
    }
}
