/**
 * Scratchy CSV class.
 */

import java.util.List;
import java.util.Iterator;

/**
 * Makes a CSV file available as a sequence of string lists.
 * The reader can handle quoted cells, but not escaped quotes within them.
 */
public class SimpleCSVReader
        implements Iterable<List<String>> {
    java.io.BufferedReader in;
    public SimpleCSVReader(java.io.Reader in) {
        this.in = new java.io.BufferedReader(in);
    }

    public Iterator<List<String>> iterator() {
        return new CSVIterator();
    }

    class CSVIterator implements Iterator<List<String>> {
        boolean atEOF = false;
        java.io.PushbackReader line;
        String oneLine = null;

        /**
         * Indicates whether there is at least one more line to come.
         * This also read in the next line if it's not currently
         * present.
         * <p>Postcondition: there is either a line in the oneLine
         * variable and we return true; or else atEOF is true and we
         * return false.
         */
        public boolean hasNext() {
            boolean oneLinePresent;
            
            if (atEOF) {
                oneLinePresent = false;
            } else {
                try {
                    if (oneLine == null) {
                        oneLine = in.readLine();
                    }
                } catch (java.io.IOException e) {
                    oneLine = null;
                }
                if (oneLine == null) {
                    atEOF = true;
                }
                oneLinePresent = oneLine != null;
            }
            assert (atEOF && !oneLinePresent) || (!atEOF && oneLinePresent);
            return oneLinePresent;
        }
            
        public List<String> next() {
            if (! hasNext()) {
                throw new java.util.NoSuchElementException("Attempt to read past end of CSV file" + in);
            }

            assert oneLine != null;

            try {
                line = new java.io.PushbackReader(new java.io.StringReader(oneLine));
                List<String> l = new java.util.ArrayList<String>();
                while (true) {
                    int c = line.read();
                    if (c < 0) {
                        // end of line
                        oneLine = null;
                        break;
                    }
                    if (c == '"') {
                        l.add(readToQuote(line));
                    } else {
                        line.unread(c);
                        l.add(readToCommaOrEnd(line));
                    }
                    c = line.read();
                    assert c == ',';
                }
                return l;
            } catch (java.io.IOException e) {
                System.err.println("Odd: uncaught IOException reading CSV file " + in);
                return null;
            }
        }

        public void remove() {
            throw new UnsupportedOperationException("SimpleCSVReader doesn't support remove()");
        }

        // Read from the current position (which will be immediately
        // after a quote character) up to the next quote character,
        // swallowing byt not including the quote character
        private String readToQuote(java.io.Reader r)
                throws java.io.IOException {
            StringBuffer sb = new StringBuffer();
            for (int c = r.read(); c != '"'; c=r.read()) {
                if (c < 0) {
                    throw new java.io.IOException("Malformed CSV file: EOF in quote");
                }
                sb.append((char)c);
            }
            return sb.toString();
        }

        // Read from the current position (which will be immediately
        // after a comma) up to the next comma or EOB.  Unread the
        // comma, or if we found the EOB, push a comma, and return the string.
        private String readToCommaOrEnd(java.io.PushbackReader r)
                throws java.io.IOException {
            StringBuffer sb = new StringBuffer();
            while (true) {
                int c = r.read();
                if (c == ',' || c < 0) {
                    r.unread(',');
                    break;
                } else {
                    sb.append((char)c);
                }
            }
            return sb.toString();
        }
    }
}
