/*
 * Strip a prepared file.
 *
 * The input file looks like this:
 *
 *     blah
 *     %%
 *     @+a,b
 *     line1
 *     @=b line2
 *     @-
 *     @+a
 *     line3
 *     @=b line4
 *     @-
 *     @+!c
 *     @!line5
 *     @=a line6
 *     @-
 *     %%
 *     more blah
 *
 * There may be leading whitespace before the @-expressions.
 *
 * The program may be called with a sequence of "-f arg" or "-f
 * arg1,arg2" options.  Each 'arg' is a string flag which may be found
 * one one of the '@+' or '@=' lines.  For each 'arg', if 'arg' is one
 * of the flags found on a '@+' line, then each line between there and
 * the subsequent '@-' line is included in the output.  If the 'arg'
 * is one of the flags in a '@=' line, then that line, minus the
 * @-spec, is included if it would appear without '@=' prefix (ie it
 * is not within a '@+'/'@-' pair which would inhibit it).  Otherwise,
 * the lines are skipped.
 *
 * The specifier !foo matches iff flag 'foo' is _not_
 * specified (when there's more than one flag, that's different from
 * saying that !foo matches all flags other than 'foo').
 *
 * '@+' trumps '@=': in the example above, if only -f b is specified,
 * 'line4' does not appear, since it is within the '@+a'/'@-' pair
 * which is suppressed by virtue of there being no '-fa' option.
 *
 * The @- lines may optionally have a spec on them as well.  If one
 * such is present, it is examined to make sure it matches the spec on
 * the @+ line it should correspond to.
 *
 * The @!key line searches for a replacement text for the key 'key',
 * and replaces this line with that text.  The replacement text is
 * provided via the '-r' option, and consists of a sequence of
 * stanzas, separated by one or more blank lines, with the first line
 * containing the key, and the remaining ones the value.  (At present)
 * only complete lines can be substituted.
 *
 * The -p option removes everything before the first, and after the
 * second, '%%', and also removes any content between {...}.  This is
 * for prettifying the input yacc source files to make the underlying
 * grammar easier to read.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/stat.h>

#define NFLAGS 4

char* progname;

typedef union {
    int size;
    struct {
        const char* key;
        const char* value;
    } cell;
} Replacement;

void Usage(void);
int strip_file(FILE*,char**,int nflags, int strip_percent, Replacement*);
int flag_match(char**flags, int nflags, char* s, int*);
void write_line(const char* line, int strip_yacc);

static Replacement* read_replacements(const char* filename);
static const char* find_replacement(const char* key, const Replacement* r);

int main(int argc, char** argv)
{
    char* flag[NFLAGS];
    int nflags = 0;
    FILE* in = stdin;
    int ch;
    int strip_percent = 0;
    Replacement* repls = NULL;

    progname = argv[0];

    while ((ch = getopt(argc, argv, "hf:pr:")) != -1) {
        switch (ch) {
          case 'h':
            Usage();
            break;
          case 'f':
            {
                char* p; char* s;
                if (nflags >= NFLAGS) {
                    fprintf(stderr, "Ooops: too many flags: maximum=%d\n", NFLAGS);
                    exit(1);
                }
                s = optarg;
                while ((p = strtok(s, ",")) != NULL) {
                    s = NULL;
                    flag[nflags] = p;
                    nflags++;
                }
            }
            break;
          case 'p':
            strip_percent = 1;
            break;
          case 'r':
            repls = read_replacements(optarg);
            break;
          default:
            Usage();
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (argc > 1) {
        Usage();
    }
    if (argc == 1) {
        if ((in = fopen(argv[0], "r")) == NULL) {
            fprintf(stderr, "Can't open file %s for input\n", argv[0]);
            exit(1);
        }
    }

    // Argument processing now complete

    strip_file(in, flag, nflags, strip_percent, repls);

    return 0;
}

/* We take the maximum line length to be MAXLINE */
#define MAXLINE 1024
struct stack_element {
    int included;
    char flagspec[MAXLINE];
};

int strip_file(FILE* in, char** flags, int nflags, int strip_yacc, Replacement* repls)
{
#define STACKSIZE 8
    char line[MAXLINE];
    struct stack_element stack[STACKSIZE];
    struct stack_element* top = &stack[0];
    int lineno = 0;
    int between_percents = 0;

    stack[0].included = 1;      /* always 1 */
    stack[0].flagspec[0] = '\0';
    while (fgets(line, MAXLINE, in) != NULL) {
        char* at;
        
        lineno++;
        if (strncmp(line, "%%", 2) == 0 && strip_yacc) {
            between_percents = !between_percents;
            continue;
        }
        if (strip_yacc && !between_percents) {
            continue;
        }

        // find the beginning of any at expression on the line, by skipping whitespace
        for (at = line; isspace(*at); at++) {
            // nothing
        }
        
        if (*at == '@') {
            switch(at[1]) {
              case '\0':        // error
                fprintf(stderr, "%s: malformed line (%d) with only @ on it\n", progname, lineno);
                break;
              case '+':
                if (top->included && flag_match(flags, nflags, &at[2], NULL)) {
                    if (top - stack == STACKSIZE) {
                        fprintf(stderr, "%s: too many nested levels on line %d\n", progname, lineno);
                        exit(1);
                    }
                    top++;
                    top->included = 1;
                } else {
                    top++;
                    top->included = 0;
                }
                strncpy(top->flagspec, &at[2], MAXLINE);
                break;
              case '-':
                if (top == stack) {
                    fprintf(stderr, "%s: more @- than @+ on line %d\n", progname, lineno);
                    exit(1);
                }
                if (! isspace(at[2])) { /* not EOL, ie there is a matching end-spec */
                    // there's a matching key on the end-line -- check this matches the corresponding start
                    if (strcmp(top->flagspec, &at[2]) != 0) {
                        fprintf(stderr, "Warning: nesting error? Expected @-%s  got %s  on line %d\n",
                                top->flagspec, at, lineno);
                    }
                }
                top--;
                break;
              case '=':
                {
                    int flaglen;
                    int strip_whitespace;
                    if (top->included
                        && (flaglen = flag_match(flags, nflags, &at[2], &strip_whitespace)) != 0) {
                        if (strip_whitespace) {
                            // avoid writing out the flag, skipping both leading or trailing whitespace
                            int linestart = (at - line) + flaglen + 2;
                            while (isspace(line[linestart])) {
                                // trim whitespace between the flag and the start of the line
                                linestart++;
                            }
                            write_line(&line[linestart], strip_yacc);
                        } else {
                            // write the full line, with the flag part blanked out
                            int i;
                            for (i=0; i<flaglen+2; i++) {
                                at[i] = ' ';
                            }
                            write_line(line, strip_yacc);
                        }
                    }
                }
                break;
              case '!':
                if (!strip_yacc && top->included) {
                    const char* repl;
                    char* p;
                    if (repls == NULL) {
                        fprintf(stderr, "%s: Found replacement specifier %s on line %d, but no -r option was present\n",
                                progname, at, lineno);
                        exit(1);
                    }

                    for (p=at; *p != '\0'; p++) {
                        // terminate at first whitespace
                        if (isspace(*p)) {
                            *p = '\0';
                            break;
                        }
                    }
                      
                    repl = find_replacement(&at[2], repls);
                    if (repl == NULL) {
                        fprintf(stderr, "%s: no replacement for key '%s' on line %d\n", progname, &at[2], lineno);
                    } else {
                        fputs(repl, stdout);
                    }
                }
                break;
              default:
                fprintf(stderr, "%s: bad character after @ on line %d\n", progname, lineno);
                break;
            }
        } else {
            if (top->included) {
                write_line(line, strip_yacc);
            }
        }
    }
    return 0;
}

void write_line(const char* line, int strip_yacc) 
{
    static int brace_level = 0;

    if (strip_yacc) {
        const char* p;
        for (p=line; *p!='\0'; p++) {
            switch (*p) {
              case '{':
                brace_level++;
                break;
              case '}':
                if (brace_level == 0) {
                    fprintf(stderr, "More close than open braces\n");
                } else {
                    brace_level--;
                }
                break;
              default:
                if (brace_level == 0) {
                    putchar(*p);
                }
                break;
            }
        }
    } else {
        fputs(line, stdout);
    }
    return;
}

/**
 * Determine whether any of the flags in flags[] match the input line,
 * which is presumed to start with an '@...' expression containing a
 * comma-separated list of strings 'arg1,arg2,!arg3,...'
 *
 * Return non-zero if any of the flags[] are present in the 'argn', or
 * if there is a '!foo' element in the @-expression, and the flag
 * 'foo' is not present in flags[].
 *
 * If strip_trailing is non-NULL, then set *strip_trailing to 1 or 0
 * depending on whether spaces after the flag should be skipped or not
 * (this currently returns 1 in all cases, but could reasonably be
 * changed so that @=foo- (say), indicated that the flag should be
 * replaced by spaces, to help with the visual appearance of the output).
 *
 * @return zero if the flag does not match, and returns
 * the length of the @-expression if it does.
 */
int flag_match(char**flags, int nflags, char* at_expression, int* strip_trailing)
{
    char* p;
    int flaglen;
    // at_expression is at most MAXLINE characters long
    char buf[MAXLINE+1];        /* +1 to ensure space for '\0' */
    char* buf_p;

    // Find the first whitespace in the string, or '\0', and copy this to buf.
    // Use strtok rather than strsep, for compatibility reasons
    // (and because its defects don't matter here)
    for (flaglen=0; at_expression[flaglen]!='\0'; flaglen++) {
        if (isspace(at_expression[flaglen])) {
            break;              /* JUMP OUT */
        }
    }
    memcpy(buf, at_expression, flaglen);
    buf[flaglen] = '\0';

    if (strip_trailing != NULL) {
        // don't use this at present -- set it to true in all cases
        *strip_trailing = 1;
    }

    buf_p = buf;
    while ((p = strtok(buf_p, ",")) != NULL) {
        int i;
        buf_p = NULL;               /* for next time */
        if (*p == '!') {
            /* This specifier '!foo' will match iff 'foo' is NOT present in flags[] */
            for (i=0; i<nflags; i++) {
                if (strcmp(&p[1], flags[i]) == 0) {
                    /* 'foo' is present: this specifier does not match */
                    return 0; /* RETURN */
                }
            }
            /* 'foo' was not present: this specifier does match */
            return flaglen;        /* RETURN */
        }
        for (i=0; i<nflags; i++) {
            if (*p == '!' && strcmp(&p[1], flags[i]) == 0) {
                return 0;    /* MATCH: RETURN */
            }
            if (strcmp(p, flags[i]) == 0) {
                /* One of the flags[] is present in this specifier -- we match */
                return flaglen;    /* RETURN */
            }
        }
    }
    return 0;
}

/*
 * The Replacement structure is a sequence of n+1 Replacement unions,
 * representing n replacements.  Replacement[0].size contains the
 * number of replacements in it, and each of the subsequent ones has a
 * (key,value) pair in Replacement[i].cell.key and .cell.value.
 */

static int cmp_replacements(const void* v1, const void* v2) 
{
    Replacement* r1 = (Replacement*)v1;
    Replacement* r2 = (Replacement*)v2;

    int res = strcmp(r1->cell.key, r2->cell.key);
    if (res == 0) {
        fprintf(stderr, "%s: WARNING: duplicate replacement keys: %s\n", progname, r1->cell.key);
    }
    return res;
}

static int cmp_replacement_to_key(const void* key, const void* v) 
{
    return strcmp((const char*)key, ((const Replacement*)v)->cell.key);
}

static Replacement* read_replacements(const char* filename) 
{
    int fd;
    Replacement* a = 0;
    int alen = 0;
    int a_allocated = 0;
    struct stat S;
    const char* file_base;
    char* buf;
    char* buf_end;
    char* p;

    if ((fd = open(filename, O_RDONLY)) < 0) {
        fprintf(stderr, "Can't open file %s to read\n", filename);
        return NULL;
    }
    if (fstat(fd, &S) != 0) {
        fprintf(stderr, "Can't stat file %s!\n", filename);
        close(fd);
        return NULL;
    }
    if ((file_base = mmap(NULL, S.st_size, PROT_READ, MAP_FILE|MAP_PRIVATE, fd, 0)) == MAP_FAILED) {
        fprintf(stderr, "Unable to map file %s\n", filename);
        close(fd);
        return NULL;
    }

    // now copy the file into a new buffer, so we can write to it
    if ((buf = (char*)malloc(S.st_size + 1)) == NULL) {
        // +1 for an extra '\0'
        fprintf(stderr, "Unable to malloc %lld bytes for file %s!\n", S.st_size, filename);
        exit(1);
    }
    memcpy(buf, file_base, S.st_size);

    // tidy up the file
    munmap((void*)file_base, S.st_size);
    file_base = 0;
    close(fd);

    p = buf;
    buf_end = buf + S.st_size;
    *buf_end = '\0';

    while (p < buf_end) {
        if (alen <= a_allocated) {
            if (a_allocated == 0) {
                // first time
                a_allocated = 8;
                if ((a = (Replacement*)malloc((a_allocated + 1) * sizeof(Replacement))) == NULL) {
                    fprintf(stderr, "Can't allocate %d replacement entries!\n", a_allocated);
                    exit(1);
                }
            } else {
                a_allocated *= 2;
                a--;            // offset back to the address that was allocated
                if ((a = (Replacement*)realloc((void*)a, (a_allocated + 1) * sizeof(Replacement))) == NULL) {
                    fprintf(stderr, "Can't reallocate %d replacement entries!\n", a_allocated);
                    exit(1);
                }
            }
            a++;                // offset, so that a[0] points to the [1]th element of the actual allocation
        }
        assert(a != NULL && a_allocated > 0 && alen < a_allocated);

        // Skip over any leading whitespace (ie at beginning of file, or extra blank lines)
        while (isspace(*p) && p < buf_end) {
            p++;
        }
        if (p == buf_end) {
            // we're at the end of file, so don't create another entry
            break;              /* JUMP OUT */
        }
        a[alen].cell.key = p;
        while (*p != '\n' && p < buf_end) {
            // skip to EOL
            p++;
        }
        if (p == buf_end) {
            break;              /* JUMP OUT */
        }
        assert(*p == '\n');
        *p = '\0';              /* end of key */

        p++;
        a[alen].cell.value = p;

        {
            int bol = 1;
            // now search forward for a blank line or end
            for (; p < buf_end; p++) {
                if (*p == '\n' && bol) {
                    break;       /* JUMP OUT */
                }
                bol = (*p == '\n');
            }
            assert(p <= buf_end);
            // Terminate the value.  If we've come to the end of the
            // file, p is now pointing one beyond the last byte in the file.
            // Bug buf is file-length+1 bytes long, so this is definitely OK.
            *p++ = '\0';
        }
        alen++;
    }

    // sort the result, in order of Replacement.cell.key
    qsort((void*)a, alen, sizeof(Replacement), &cmp_replacements);

    a--;                        /* offset back to the beginning of the actual allocation */
    a[0].size = alen;
    
    return a;
}

static const char* find_replacement(const char* key, const Replacement* r)
{
    int nreplacements = r->size;
    Replacement* match = (Replacement*)bsearch((const void*)key,
                                               (const void*)&r[1],
                                               nreplacements,
                                               sizeof(Replacement),
                                               &cmp_replacement_to_key);
    return (match == NULL ? NULL : match->cell.value);
}

void Usage(void) 
{
    fprintf(stderr, "Usage: %s [-p] [-f flag[,flag]]... [-r replacement-text] [inputfile]\n", progname);
    exit(1);
}
