/*
 * This module performs the tests described in ../../grammar/testcases.csv. 
 *
 * Usage: java -cp ... ParseAll -t -pxxx ../../grammar/testcases-xxx.csv
 *
 * Here, 'xxx' is one of the supported parsers: fits, ogip or cds.
 */

import uk.me.nxg.unity.*;

import java.util.List;

public class ParseAll 
{
    public static void main(String[] args) 
    {
        final int LIST_LEXEMES = 1;
        final int DISPLAY_PARSE = 2;
        final int RUN_TESTS = 3;
        int action = 0;
        
        String inputString = null;
        char optionChar;
        String parser = "fits";

        for (int i=0; i<args.length; i++) {
            if (args[i].charAt(0) == '-') {
                switch (args[i].charAt(1)) {
                  case 'l':
                    action = LIST_LEXEMES;
                    break;
                  case 'd':
                    action = DISPLAY_PARSE;
                    break;
                  case 't':
                    action = RUN_TESTS;
                    break;
                  case 'p':
                    parser = args[i].substring(2);
                    break;
                  case 'S':
                    {
                        java.util.Set<String> parsers = UnitParser.getParserTypes();
                        for (String s : parsers) {
                            System.out.println("parser: " + s);
                        }
                    }
                    break;
                  default:
                    Usage();
                    break;
                }
            } else {
                inputString = args[i];
            }
        }

        if (inputString == null || action == 0) {
            Usage();
        }

        // if (! UnitParser.getParserTypes().contains(parser)) {
        //     System.err.println(""

        try {
            switch (action) {
              case LIST_LEXEMES:
                listLexemes(inputString);
                break;
              case DISPLAY_PARSE:
                // parse the input string and display the contents in an unambiguous way
                displayParse(inputString, parser);
                break;
              case RUN_TESTS:
                runTestsInFile(inputString, parser);
                break;
              default:
                Usage();
            }
        } catch (UnitParserException e) {
            System.err.println("Unit parser exception: " + e);
        }
    }

    /**
     * List lexemes in the inputString, for debugging 
     */
    private static void listLexemes(String inputString)
            throws UnitParserException {
        UnitParser p = new UnitParser("fits", inputString); // the parser type doesn't matter
        UnitParser.Lexeme lexeme;
        while ((lexeme = p.getLexeme()) != null) {
            System.out.println("Token: " + lexeme.token + ", yylval=" + lexeme.value);
        }
    }

    /**
     * Display the parsed version of the input string in an unambiguous way
     */
    private static void displayParse(String inputString, String parser) {
        try {
            UnitExpr units = new UnitParser(parser, inputString).getParsed();
            System.out.print(units + ":");
            List<OneUnit> lu = units.asList();
            for (OneUnit u : lu) {
                System.out.print("  " + u.toDebugString());
            }
            System.out.println();
        } catch (UnitParserException e) {
            System.err.println("Failed to parse that: " + e);
        }
    }

    /**
     * Run the tests in the named CSV file
     */
    private static void runTestsInFile(String filename, String parser)
            throws UnitParserException {
        try {
            java.io.BufferedReader in =new java.io.BufferedReader(new java.io.FileReader(filename));
            String line;
            int nok = 0;
            int nfail = 0;

            // Set the locale to something non-default, so we can
            // check that the code is locale-independent.  This isn't
            // a very thorough test, because the French locale seems
            // to differ from the non-locale only in the use of commas
            // for the decimal separator.  However I can't find a
            // locale which differs more than this in relevant
            // respects -- all the ones I would have expected to be
            // different (japaense, chinese, arabic, thai, ...) seem to have
            // 'us' digits (pragmatically, I suppose).
            java.util.Locale.setDefault(new java.util.Locale("FR"));

            while ((line = in.readLine()) != null) {
                if (line.length() > 0 && line.charAt(0) != '#') {
                    String[] fields = line.split(",[ \\t]*");

                    String unitString;
                    boolean shouldBeOK;
                    String failureReason;
                    String parseResult;
                    String expected = null;
                    UnitExpr expr;

                    // expectedStatus indicates further tests:
                    //   0: no further tests
                    //   1: includes unrecognised units
                    //   2: all units recognised, but at least one deprecated
                    //   3: uses SI prefixes where forbidden
                    //   4: all constraints satisfied
                    int expectedStatus = 0;

                    // column 0
                    {
                        unitString = fields[0];
                        StringBuffer sb = new StringBuffer();
                        try {
                            expr = new UnitParser(parser, fields[0]).getParsed();
                            parseResult = expr.toString("debug");
                        } catch (UnitParserException e) {
                            // that's OK -- silently acknowledge this
                            parseResult = null;
                            expr = null;
                        }
                    }

                    // column 1
                    shouldBeOK = (fields[1].length() == 0);
                    failureReason = fields[1];

                    // column 2
                    if (fields.length > 2) {
                        char leadingChar = fields[2].charAt(0);
                        if (Character.isDigit(leadingChar) || leadingChar == '-') {
                            // OK
                            expectedStatus = 0;
                            expected = fields[2];
                        } else {
                            expected = fields[2].substring(1);
                            switch (leadingChar) {
                              case 'x':
                                expectedStatus = 1;
                                break;
                              case '?':
                                expectedStatus = 2;
                                break;
                              case 's':
                                expectedStatus = 3;
                                break;
                              case '!':
                                expectedStatus = 4;
                                break;
                              default:
                                // oops -- the above are the only allowed flags
                                System.err.println("Malformed test: unrecognised flag in " + fields[2]);
                                System.exit(1);
                            }
                        }
                    }

                    // we've gathered all the information -- now do the check

                    if (shouldBeOK) {
                        if (parseResult == null) {
                            System.out.println("Parsing \"" + unitString
                                               + "\" should have worked, but didn't");
                            nfail++;
                        } else if (parseResult.equals(expected)) {
                            // this test passed
                            nok++;
                            switch (expectedStatus) {
                              case 0:
                                break; // no other checks required
                              case 1:
                                // this string had non-recognised units
                                if (expr.allUnitsRecognised(parser)) {
                                    System.out.println("Parsing \"" + unitString
                                                       + "\"(" + parser + ") has non-recognised units: didn't spot them!");
                                    nfail++;
                                } else {
                                    nok++;
                                }
                                break;
                              case 2:
                                // all units recognised, but at least one deprecated
                                if (expr.allUnitsRecognised(parser) && !expr.allUnitsRecommended(parser)) {
                                    nok++;
                                } else {
                                    System.out.println("Parsing \"" + unitString
                                                       + "\"(" + parser + ") has deprecated units: didn't spot them");
                                    nfail++;
                                }
                                break;
                              case 3:
                                // at least one unit was used with an SI prefix, which
                                // is recognised and known not to permit this
                                if (expr.allUsageConstraintsSatisfied(parser)) {
                                    System.out.println("Parsing \"" + unitString
                                                       + "\"(" + parser + ") does not conform to usage constraints: didn't spot that");
                                    nfail++;
                                } else {
                                    nok++;
                                }
                                break;
                              case 4:
                                // fully conformant
                                if (expr.isFullyConformant(parser)) {
                                    nok++;
                                } else {
                                    System.out.println("Parsing \"" + unitString
                                                       + "\"(" + parser + ") is fully conformant: didn't spot that");
                                    nfail++;
                                }
                                break;
                              default:
                                throw new AssertionError("Impossible expectedStatus: " + expectedStatus);
                            }
                        } else {
                            System.out.println("Parsing \"" + unitString
                                               + "\": expected \""  + expected
                                               + "\" got \"" + parseResult + "\"");
                            nfail++;
                        }
                        // separately, check the formatted versions, if they're present
                        if (expr != null && fields.length > 3) {
                            // FIXME: should check that reparsing the formatted produces the same expr
                            // (define UnitExpr.equals(UnitExpr) method?)
                            for (int i=3; i<fields.length; i++) {
                                String[] parts = fields[i].split(":");
                                if (parts.length != 2) {
                                    System.err.println("Malformed formatting specifier: " + fields[i]);
                                } else {
                                    String actualFormat = expr.toString(parts[0]);
                                    String expectedFormat = parts[1];
                                    if (actualFormat.equals(expectedFormat)) {
                                        nok++;
                                    } else {
                                        System.out.println("Formatting (" + parts[0]
                                                           + "): expected \"" + expectedFormat
                                                           + "\", got \"" + actualFormat + "\"");
                                        nfail++;
                                    }
                                }
                            }
                        }
                    } else {
                        // should have failed
                        if (parseResult == null) {
                            // this test was expected to fail
                            nok++;
                        } else {
                            System.out.println("Parsing \"" + unitString
                                               + "\" should have failed ("
                                               + failureReason
                                               + "), but it produced: "
                                               + parseResult);
                            nfail++;
                        }
                    }
                }
            }
            StringBuilder sb = new StringBuilder("      ");
            sb.replace(sb.length()-parser.length(), sb.length(), parser);
            if (nfail == 0) {
                System.out.println(sb + " : ALL " + nok + " tests pass");
            } else {
                System.out.println(sb + " : passes: " + nok);
                System.out.println("         FAILS:  " + nfail);
            }
        } catch (java.io.IOException e) {
            System.err.println("IOException reading " + filename + ": " + e);
        }
    }
    
    private static void Usage() {
        System.err.println("Usage: ParseAll -lpt unitstring");
        System.exit(1);
    }
}
