/*
 * Supports various other tests not covered by the tests in
 * ../../grammar/testcases.csv.  These are generally tests of the
 * java-specific parts of the interface, rather than tests of what
 * does and doesn't parse.
 */

import uk.me.nxg.unity.*;

import org.junit.Test;
import static org.junit.Assert.*;

public class TestMisc {

    @Test public void testFitsParse() throws Exception {
        UnitParser p = new UnitParser("fits", "mm^3/s");
        UnitExpr expr = p.getParsed();
        assertTrue(expr.allUnitsRecognised("fits"));

        // the leading factor is 1.0; log(1.0)=0.0
        assertEquals("logFactor", 0.0, expr.getLogFactor(), 0);

        // There are no Cd in this unit
        OneUnit u = expr.getUnit("Cd");
        assertNull(u);

        u = expr.getUnit("m");
        assertEquals("m", u.getBaseUnit());
        assertEquals("m-prefix", -3, u.getPrefix());
        assertEquals("m-power", 3.0, u.getPower(), 0);

        u = expr.getUnit("s");
        assertEquals("s", u.getBaseUnit());
        assertEquals("s-prefix", 0, u.getPrefix());
        assertEquals("s-power", -1.0, u.getPower(), 0);
    }

    @Test public void testErgs() throws Exception {
        // ergs are
        //   permitted but deprecated in fits
        //   permitted and fine in ogip, and 
        //   unrecognised in cds
        UnitExpr expr = new UnitParser("fits", "erg").getParsed();
        assertTrue(expr.allUnitsRecognised("fits"));
        assertTrue(expr.allUnitsRecognised("ogip"));
        assertFalse(expr.allUnitsRecognised("cds"));

        OneUnit erg = expr.getUnit("erg");
        assertTrue(erg.isRecognisedUnit("fits"));
        assertFalse(erg.isRecommendedUnit("fits"));
        assertTrue(erg.isRecognisedUnit("ogip"));
        assertTrue(erg.isRecommendedUnit("ogip"));
        assertFalse(erg.isRecognisedUnit("cds"));
        assertFalse(erg.isRecommendedUnit("cds"));
        
        assertEquals("erg", erg.getBaseUnit());
        assertEquals("erg", erg.getBaseUnit("cds"));
    }

    @Test public void preferredUnits() throws Exception {
        // Check recommended/deprecated status, and preferred output forms
        // FITS: accepts pixel and pix, prefers pixel
        // OGIP: accepts pixel
        // CDS:  accepts pix
        UnitExpr expr = new UnitParser("fits", "pix").getParsed();

        OneUnit pix = expr.getUnit("pix");
        assertNotNull(pix);
        assertEquals("pixel", pix.getBaseUnit("fits"));
        assertEquals("pixel", pix.getBaseUnit("ogip"));
        assertEquals("pix",   pix.getBaseUnit("cds"));
        assertEquals("pixel", pix.getBaseUnit());
        
        OneUnit pixel = expr.getUnit("pixel");
        assertNotNull(pixel);
    }

    @Test public void unusualUnits() throws Exception {
        // parse femto-furlong (just "furlong" gets parsed as femto-urlong)
        UnitExpr expr = new UnitParser("fits", "ffurlong").getParsed();
        OneUnit furlong = expr.getUnit("furlong");
        assertNotNull(furlong);
        assertEquals("furlong", furlong.getBaseUnit());
    }

    @Test public void outputInLocale() throws Exception {
        UnitExpr expr = new UnitParser(UnitParser.CDS, "0.1nm/s").getParsed();
        String cdsFormatDefault = expr.toString(UnitParser.CDS);
        String cdsFormatFR = expr.toString(UnitParser.CDS, java.util.Locale.FRENCH);
        assertEquals("0.1nm/s", cdsFormatDefault);
        assertEquals("0,1nm/s", cdsFormatFR);
    }

    public static void main(String args[]) {
        org.junit.runner.JUnitCore.main("TestMisc");
    }
}
