package uk.me.nxg.unity;

import java.util.List;


/**
 * A single unit.  For example, the string "m s^-2" will result in two
 * OneUnit instances, corresponding to "m" and "s^-2".
 */
public class OneUnit {
    private int prefixPower;
    private String baseUnitString;
    private UnitDefinition baseUnitDefinition;
    private double exponent;

    /** 
     * A regexp to match a unit with an SI prefix.
     * The order in this regexp is important -- the "da" must match
     * before "d" can, so that "dam" is parsed as "da"+"m", and not
     * "d"+"am".  Also, this regexp requires there to be a unit: "m"
     * is not parsed as a prefix plus an empty unit.
     */
    private static final java.util.regex.Pattern unitWithPrefix
            = java.util.regex.Pattern.compile("(da|[yzafpnumcdhkMGTPEZY])(.+)");

    /**
     * Private internal constructor
     */
    private OneUnit(int prefix, String us, UnitDefinition ud, double e) {
        prefixPower = prefix;
        if (us == null) {
            baseUnitDefinition = ud;
        } else {
            baseUnitString = us;
        }
        assert (baseUnitString == null && baseUnitDefinition != null) || (baseUnitString != null && baseUnitDefinition == null);
        exponent = e;
    }

    /**
     * Constructer of base unit -- no multiple.
     */
    OneUnit(String s, int e) {
        this(s, (double)e);
    }

    /**
     * Constructer of base unit
     * @param unitString a string representation of a single unit with
     * a possible SI prefix
     * @param e the power to which this unit is raised
     */
    OneUnit(String unitString, double e) {
        // It would be possible/appropriate to do strict checking of
        // units in here -- that is, throwing exceptions whenever we
        // encounter any non-recommended unit (ie one which lookupUnit
        // returns null for).  This would create the additional
        // possibility of being a convenient/reasonable place to check
        // whether the abbreviation that we look up is an allowed one
        // for the syntax we're parsing, as long as we were
        // additionally provided with some information within this
        // constructor, about which syntax that is (which is currently
        // not available in here).
        UnitDefinition def;
        if ((def = lookupUnit(unitString)) != null) {
            prefixPower = 0;
            baseUnitDefinition = def;
        } else {
            java.util.regex.Matcher m = unitWithPrefix.matcher(unitString);
            if (m.matches()) {
                prefixPower = prefixLetterToPower(m.group(1).equals("da")
                                                  ? UnitVal.CHAR_DECA
                                                  : m.group(1).charAt(0));
                def = lookupUnit(m.group(2));
                if (def == null) {
                    baseUnitString = m.group(2);
                } else {
                    baseUnitDefinition = def;
                }
            } else {
                prefixPower = 0;
                baseUnitString = unitString;
            }
        }
        assert (baseUnitString == null && baseUnitDefinition != null) || (baseUnitString != null && baseUnitDefinition == null);
        exponent = e;
    }

    /**
     * Produce the reciprocal of the unit -- negate the power.
     * @return this
     */
    OneUnit reciprocate() {
        exponent *= -1;
        return this;
    }

    /**
     * Returns the prefix of the unit, as a base-ten log.  Thus a
     * prefix of "m", for "milli", would produce a prefix of -3.
     */
    public int getPrefix() {
        return prefixPower;
    }

    /**
     * Returns the power to which this unit is raised.
     * A unit of "mm^3" would return 3.0, and "m^(-1/2)" would return -0.5.
     */
    public double getPower() {
        return exponent;
    }    
    
    /**
     * Convert SI prefixes to the decimal power they correspond to
     */
    private static int prefixLetterToPower(char prefixLetter) {
        int prefixPower;
        switch (prefixLetter) {
          case 'y': prefixPower = -24; break;
          case 'z': prefixPower = -21; break;
          case 'a': prefixPower = -18; break;
          case 'f': prefixPower = -15; break;
          case 'p': prefixPower = -12; break;
          case 'n': prefixPower =  -9; break;
          case 'u': prefixPower =  -6; break;
          case 'm': prefixPower =  -3; break;
          case 'c': prefixPower =  -2; break;
          case 'd': prefixPower =  -1; break;
          case UnitVal.CHAR_DECA:
            prefixPower = 1;
            break;
          case 'h': prefixPower =   2; break;
          case 'k': prefixPower =   3; break;
          case 'M': prefixPower =   6; break;
          case 'G': prefixPower =   9; break;
          case 'T': prefixPower =  12; break;
          case 'P': prefixPower =  15; break;
          case 'E': prefixPower =  18; break;
          case 'Z': prefixPower =  21; break;
          case 'Y': prefixPower =  24; break;
          default:
            // The lexer in lex.jflex should make this impossible
            throw new AssertionError("Impossible prefix: '" + prefixLetter + "'");
        }
        return prefixPower;
    }

    /**
     * Convert a prefix power to a String.  These are returned as
     * Strings, so that we can later consider 'da' or 'ki' (etc) as prefixes.
     */
    private static String prefixPowerToString(int prefixPower) {
        String rval;
        switch (prefixPower) {
          case -24: rval = "y"; break;
          case -21: rval = "z"; break;
          case -18: rval = "a"; break;
          case -15: rval = "f"; break;
          case -12: rval = "p"; break;
          case  -9: rval = "n"; break;
          case  -6: rval = "u"; break;
          case  -3: rval = "m"; break;
          case  -2: rval = "c"; break;
          case  -1: rval = "d"; break;
          case   0: rval = "";  break;
          case   1: rval = "da";break;
          case   2: rval = "h"; break;
          case   3: rval = "k"; break;
          case   6: rval = "M"; break;
          case   9: rval = "G"; break;
          case  12: rval = "T"; break;
          case  15: rval = "P"; break;
          case  18: rval = "E"; break;
          case  21: rval = "Z"; break;
          case  24: rval = "Y"; break;
          default:
            // This should be impossible, since it's only
            // prefixLetterToPower which sets these values
            throw new AssertionError("Impossible prefix power: " + prefixPower);
        }
        return rval;
    }

    /**
     * Convert a prefix power to a String, in a form useful within
     * siunitx.  The only real difference from {@link
     * prefixPowerToString} is that <code>\micro</code> is formatted differently.
     */
    private static String prefixPowerToStringLaTeX(int prefixPower) {
        String rval;
        switch (prefixPower) {
          case -24: rval = "\\yocto"; break;
          case -21: rval = "\\zepto"; break;
          case -18: rval = "\\atto";  break;
          case -15: rval = "\\femto"; break;
          case -12: rval = "\\pico";  break;
          case  -9: rval = "\\nano";  break;
          case  -6: rval = "\\micro"; break;
          case  -3: rval = "\\milli"; break;
          case  -2: rval = "\\centi"; break;
          case  -1: rval = "\\deci";  break;
          case   0: rval = "";        break;
          case   1: rval = "\\deca";  break;
          case   2: rval = "\\hecto"; break;
          case   3: rval = "\\kilo";  break;
          case   6: rval = "\\mega";  break;
          case   9: rval = "\\giga";  break;
          case  12: rval = "\\tera";  break;
          case  15: rval = "\\peta";  break;
          case  18: rval = "\\exa";   break;
          case  21: rval = "\\zetta"; break;
          case  24: rval = "\\yotta"; break;
          default:
            // This should be impossible, since it's only
            // prefixLetterToPower which sets these values
            throw new AssertionError("Impossible prefix power: " + prefixPower);
        }
        return rval;
    }

    static List<OneUnit> divide(List<OneUnit> num, List<OneUnit> den) {
        List<OneUnit> res = new java.util.ArrayList<OneUnit>(num);
        for (OneUnit u : den) {
            res.add(new OneUnit(u.prefixPower, u.baseUnitString, u.baseUnitDefinition, -u.exponent));
        }
        return res;
    }

    /**
     * Format this unit in some sort of canonical form.  The form is
     * unspecified.  This should
     * not generally be used for formatted output -- for that, it will
     * generally be more appropriate to use {@link UnitExpr#toString}.
     */
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(unitString());
        if (exponent != 1) {
            sb.append('^').append(exponent);
        }
        return sb.toString();
    }

    /**
     * Obtains the string form of the unit, including prefix, in the given syntax.  That is,
     * return 'mm' not 'm'.
     */
    public String unitString(String syntaxName) {
        if (syntaxName.equals(UnitParser.LATEX)) {
            return unitStringLaTeX(syntaxName);
        } else {
            return unitStringText(syntaxName);
        }
    }

    public UnitDefinition getBaseUnitDefinition() {
        UnitDefinition d;
        if (baseUnitDefinition != null) {
            d = baseUnitDefinition;
        } else {
            assert baseUnitString != null;
            d = UnitDefinitionMap.getInstance().lookup(baseUnitString);
        }
        return d;
    }

    /**
     * Returns the base unit for this string, in a default syntax
     * @return the base unit as, for example, "m" for metres
     */
    public String getBaseUnit() {
        return getBaseUnit(null);
    }

    /**
     * Returns the base unit for this string, in the given syntax
     * @param syntaxName one of the syntaxes of {@link UnitParser}
     * @return the base unit as, for example, "m" for metres
     */
    public String getBaseUnit(String syntaxName) {
        String baseUnitAsString;
        if (baseUnitString == null) {
            UnitDefinition.Syntax stx = baseUnitDefinition.getSyntax(syntaxName);
            if (stx == null) {
                // This is a non-recommended unit for this syntax, so
                // there's no preferred form for this unit
                stx = baseUnitDefinition.getSyntax(null);
            }
            assert stx != null;
            baseUnitAsString = stx.getOneAbbrev();
        } else {
            baseUnitAsString = baseUnitString;
        }
        assert baseUnitAsString != null;
        return baseUnitAsString;
    }
    
    /**
     * Retrieve the string form of the unit in one of the text syntaxes.
     * @param syntaxName one of the syntaxes of {@link UnitParser}
     * @return the text unit as, for example, "mm" for millimetres
     */
    private String unitStringText(String syntaxName) {
        StringBuilder sb = new StringBuilder();
        if (prefixPower != 0) {
            sb.append(prefixPowerToString(prefixPower));
        }
        sb.append(getBaseUnit(syntaxName));
        return sb.toString();
    }

    /**
     * Retrieve the string form of the unit in a form suitable for use within the siunitx LaTeX package.
     * The LaTeX serialisation isn't quite a syntax like the text ones,
     * though it appears similar to one from the outside.
     */
    private String unitStringLaTeX(String syntaxName) {
        StringBuilder sb = new StringBuilder();
        if (prefixPower != 0) {
            sb.append(prefixPowerToStringLaTeX(prefixPower)).append(' ');
        }
        String printUnit;
        if (baseUnitString == null) {
            UnitDefinition.Syntax stx = baseUnitDefinition.getSyntax(syntaxName);
            if (stx == null) {
                // This is a non-recommended unit for this syntax, so
                // there's no preferred form for this unit
                stx = baseUnitDefinition.getSyntax(null);
            }
            printUnit = stx.latexAbbrev();
        } else {
            printUnit = baseUnitString;
        }
        assert printUnit != null && ! (printUnit.length()==0);
        sb.append(printUnit);

        return sb.toString();
    }

    /**
     * Obtains the string form of the unit, including prefix, with a default syntax.
     * That is, return 'mm' not 'm'.
     */
    public String unitString() {
        return unitString(UnitParser.FITS);
    }
    
    /**
     * Obtains the power the unit is raised to.  For example, for the
     * unit 'mm^2', return 2.
     */
    public double getExponent() {
        return exponent;
    }

    /**
     * Indicates whether the base unit is one of those recognised
     * within the specification of the given syntax.  In this context,
     * "recognised" means "mentioned in the specification", so deprecated units
     * count as recognised ones.
     *
     * <p>Note that this checks that the <em>unit</em> is a
     * recommended one: we don't (currently) check whether the
     * abbreviation that got us here is a recommended one (for
     * example, "pixel" is a valid FITS/CDS name for pixels, and "pix"
     * is a FITS and OGIP one).
     * @param syntax one of the syntaxes of {@link UnitParser}
     * @see #isRecommendedUnit
     */
    public boolean isRecognisedUnit(String syntax) {
        // We could check the validity of the abbreviation whilst we're
        // parsing the units, but that would frustrate our ability to
        // subsequently write out recommended units in the appropriate
        // form.
        return baseUnitDefinition != null
                && baseUnitDefinition.getSyntax(syntax) != null;
    }

    /**
     * Indicates whether the base unit is one of those recommended
     * within the specification of the given syntax.  In this context,
     * "recommended" means "mentioned in the specification"
     * <em>and</em> not deprecated.
     *
     * <p>Note that this checks that the <em>unit</em> is a
     * recommended one: we don't (currently) check whether the
     * abbreviation that got us here is a recommended one (for
     * example, "pixel" is a valid FITS/CDS name for pixels, and "pix"
     * is a FITS and OGIP one).
     * @param syntax one of the syntaxes of {@link UnitParser}
     * @see #isRecognisedUnit
     */
    public boolean isRecommendedUnit(String syntax) {
        // We could check the validity of the abbreviation, when we're
        // parsing the units, but that would frustrate our ability to
        // subsequently write out recommended units in the appropriate
        // form.
        if (baseUnitDefinition == null) {
            return false;
        } else {
            UnitDefinition.Syntax s = baseUnitDefinition.getSyntax(syntax);
            return s != null && !s.isDeprecated();
        }
    }

    /**
     * Indicates whether the unit is being used in a way which
     * satisfies any usage constraints.  Principally, this tests
     * whether a unit which may not be used with SI prefixes was
     * provided with a prefix, but there may be other constraints
     * present.
     *
     * <p>An unrecognised unit has no constraints, and so will always
     * satisfy them; this extends to units which are unrecognised in a
     * particular syntax.
     * @param syntax one of the syntaxes of {@link UnitParser}
     */
    public boolean satisfiesUsageConstraints(String syntax) {
        // Other constraints possible: for some syntaxes, the 'crab'
        // unit may be used with only 'm' prefix.
        boolean validity;
        if (baseUnitDefinition == null) {
            // there are no constraints, so this necessarily satisfies all of them
            validity = true;
        } else {
            UnitDefinition.Syntax s = baseUnitDefinition.getSyntax(syntax);
            if (s == null) {
                // as above
                validity = true;
            } else {
                validity = prefixPower == 0 || s.mayHaveSIPrefixes();
            }
        }
        return validity;
    }
    

    // /**
    //  * Indicates whether the base unit is recommended but deprecated
    //  * within the specification of the given syntax.  That is, an
    //  * <em>unknown</em> unit is not deprecated in this sense.
    //  * @param syntax one of the syntaxes of {@link UnitParser}
    //  * @see #isRecommendedUnit
    //  */
    // public boolean isDeprecatedUnit(String syntax) {
    //     if (baseUnitDefinition == null) {
    //         return false;
    //     } else {
    //         UnitDefinition.Syntax s = baseUnitDefinition.getSyntax(syntax);
    //         return s != null && s.isDeprecated();
    //     }
    // }

    /**
     * Write out the unit in a testable format.
     * This is for debugging and testing.
     */
    public String toDebugString() {
        StringBuilder sb = new StringBuilder();
        sb.append(prefixPower).append('/');
        if (baseUnitString == null) {
            sb.append(baseUnitDefinition.getSyntax(UnitParser.FITS).getOneAbbrev());
        } else {
            sb.append(baseUnitString);
        }
        sb.append('/');
        if (exponent == Math.floor(exponent)) { // no fractional part
            sb.append(Math.round(exponent)); // print an integer
        } else {
            sb.append(exponent);
        }
        return sb.toString();
    }

    // Helper methods
    private UnitDefinition lookupUnit(String unitString) {
        return UnitDefinitionMap.getInstance().lookup(unitString);
    }
}
