package uk.me.nxg.unity;

import java.util.List;

/**
 * Describes a unit.
 *
 * <p>A 'unit' is a notion like 'metre', or 'pixel', and is not
 * dependent on a particular syntax.  The syntax-specific aspects of
 * this (described by {@link UnitDefinition.Syntax}) are how the unit
 * is abbreviated, and indeed whether it is permitted in that syntax.
 */
public class UnitDefinition implements java.io.Serializable {
    private String name;    // the name of the unit, such as 'metre' -- not the abbreviation 'm'
    private String type;    // type such as 'length'
    private String notes;   // for example 'Hz = s-1'
    private String latexForm;   // special LaTeX form, if different from abbrev
    private java.util.Map<String,UnitDefinition.Syntax> syntaxes = null;

    // The following is unlikely to change, but its presence keeps the compiler happy
    private static final long serialVersionUID = 42L;

    public UnitDefinition(String name, String type, String notes, String latexForm) {
        this.name = name;
        this.type = type;
        this.notes = notes;
        this.latexForm = (latexForm.length()==0) ? null : latexForm;
        syntaxes = new java.util.HashMap<String,UnitDefinition.Syntax>(5);
        // This map will have three keys, and 5 > 3/(default load factor=0.75)
    }

    // accessors
    /** The name of this unit, for example 'metre' */
    public String name()  { return name; }
    /** A description of the type of this unit, for example 'length' */
    public String type()  { return type; }
    /** Further remarks about this unit, or other comments */
    public String notes() { return notes; }
    /** A LaTeX version of the unit abbrev, or null if a
     * syntax-specific abbrev should be used
     */
    public String latexForm() { return latexForm; }
    
    /**
     * Return the syntax-specific information about this unit.
     * Returns null if the syntax is unknown for this unit, meaning
     * that the given syntax does not recognise this unit as a
     * recommended one.
     *
     * <p>As a special case, if the syntax is given as
     * <code>null</code>, then return <em>any</em> syntax (the
     * caller is presumably scratching around for something to print).
     * @param syntax a string name for the syntax, which should be one
     * of the syntaxes of {@link UnitParser}, or null if we want any
     * syntax at all
     * @return the syntax details for this unit in this syntax, or null if the syntax is unknown
     */
    public Syntax getSyntax(String syntax) {
        if (syntax == null) {
            java.util.Set<String> keys = syntaxes.keySet();
            assert !keys.isEmpty();
            String oneKey = keys.iterator().next();
            return syntaxes.get(oneKey);
        } else {
            return syntaxes.get(syntax);
        }
    }

    /**
     * Produces a set of the syntax names for this unit
     */
    public java.util.Set<String> getSyntaxes() {
        return syntaxes.keySet();
    }

    /**
     * Produces a representation of this unit as a string.  This
     * should not generally be used for formatting expressions: for that, use
     * {@link UnitExpr#toString}.
     */
    public String toString() {
        return name + "(" + type + ")";
    }

    // mutators

    /**
     * Add a new syntax to this unit definition.
     * The string describing the syntax can contain the characters
     * '1', 's' or 'd' to indicate respectively that the syntax is
     * permissible, that SI prefixes may be used with it, and that
     * it is nonetheless deprecated.  There may be more than one
     * permissible name for a unit (for example, "pix" and "pixel"
     * may both be legitimate names for pixels).
     *
     * <p>This is not generally useful for clients --
     * it is intended for use while building the package.
     *
     * @param syntax the syntax being updated (eg "fits")
     * @param abbrev the permissible name for this unit in this syntax (eg "m" for metre)
     * @param syntaxSpec a string describing the syntax
     * @throws UnitParserException if the syntax is not a recognised one
     */
    public boolean addSyntax(String syntax, String abbrev, String syntaxSpec)
            throws UnitParserException {
        if (! SyntaxFactory.isKnownParser(syntax)) {
            throw new UnitParserException("Unrecognised syntax \"" + syntax
                                          + "\": recognised types are " + SyntaxFactory.getParserTypes());
        }
        if (syntaxSpec.contains("1")) {
            Syntax s = syntaxes.get(syntax);
            if (s == null) {
                syntaxes.put(syntax,
                             new Syntax(syntax,
                                        syntaxSpec.contains("s"),
                                        syntaxSpec.contains("d"),
                                        abbrev));
            } else {
                s.addAbbrev(abbrev);
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * Describes the per-syntax information about a unit.
     */
    public class Syntax implements java.io.Serializable {
        private String syntaxName; // the name of the syntax being described (eg "fits")
        private List<String> abbrevs; // such as 'm' for metre (typically just one member)
        private boolean mayHaveSIPrefixes;
        private boolean isDeprecated;

        private static final long serialVersionUID = 43L;

        Syntax(String syntaxName, boolean mayHaveSIPrefixes, boolean isDeprecated, String firstAbbrev) {
            if (syntaxName == null) {
                throw new IllegalArgumentException("Syntax must have a non-null syntax name");
            }
            this.syntaxName = syntaxName;
            this.mayHaveSIPrefixes = mayHaveSIPrefixes;
            this.isDeprecated = isDeprecated;
            this.abbrevs = new java.util.ArrayList<String>(1);
            if (firstAbbrev == null) {
                throw new IllegalArgumentException("Syntax abbreviations must be non-null");
            }
            addAbbrev(firstAbbrev);
            assert syntaxName != null && abbrevs.size() >= 1;
        }

        void addAbbrev(String abbrev) {
            abbrevs.add(abbrev);
        }

        /**
         * Indicates how this unit is described in this syntax.  For
         * example, the unit 'metre' will typically be representable
         * by only 'm', in which case the return value would be
         * <code>["m"]</code>.
         * @return a list of permitted string representations of this
         * unit in this syntax (usually just one member, but more are permissable)
         */
        public List<String> getAbbrevs() {
            return abbrevs;
        }

        /**
         * Get a single abbrev for this unit.  If there is more than
         * one allowable for this syntax, it is unspecified which one
         * is returned.
         * @return an abbreviation for this unit (there will always be
         * at least one)
         */
        public String getOneAbbrev() {
            List<String> abbrevs = getAbbrevs();
            assert abbrevs.size() >= 1;
            return abbrevs.get(0);
        }

        /**
         * Get a version of an abbrev, suitable for outputting as part
         * of the LaTeX representation of this unit.  This will
         * generally be the unit abbrev, unless this has been
         * overridden.
         * @return an abbrev, which is non-empty
         */
        public String latexAbbrev() {
            String rval;
            if (latexForm() == null) {
                rval =  getOneAbbrev();
            } else {
                rval = latexForm();
            }
            assert rval != null && rval.length() != 0;
            return rval;
        }

        /**
         * Indicates whether this unit may be used with SI prefixes.
         */
        public boolean mayHaveSIPrefixes() {
            return mayHaveSIPrefixes;
        }

        /**
         * Indicates whether this syntax deprecates (but still
         * permits) the use of this unit.
         */
        public boolean isDeprecated() {
            return isDeprecated;
        }

        public String toString() {
            StringBuilder sb = new StringBuilder(syntaxName);
            sb.append(' ');
            sb.append(mayHaveSIPrefixes() ? "s" : "-").append(isDeprecated() ? "d" : "-");
            sb.append('>');
            String sep = "";
            for (String pfx : getAbbrevs()) {
                sb.append(sep);
                sep = "/";
                sb.append(pfx);
            }
            return sb.toString();
        }
    }
}
