package uk.me.nxg.unity;

import java.util.List;

/**
 * A parsed unit expression.
 */
public class UnitExpr {
    List<OneUnit> unitsList;
    double logFactor;
    
    /**
     * Construct a UnitExpr instance from a single list of units
     * @param logFactor the base-ten log of the factor at the head of the unit specification
     * @param l the list of units
     */
    UnitExpr(double logFactor, List<OneUnit> l) {
        this.logFactor = logFactor;
        unitsList = l;
    }

    /**
     * Construct a UnitExpr instance from a pair of lists of units.
     * Divide the numerator by the denominator.
     * @param logFactor the base-ten log of the factor at the head of the unit specification
     * @param num the list of units in the numerator
     * @param den the list of units in the numerator
     */
    UnitExpr(double logFactor, List<OneUnit> num, List<OneUnit> den) {
        this.logFactor = logFactor;
        unitsList = OneUnit.divide(num, den);
    }

    /**
     * Obtain the base-10 log of the factor multiplying this expression
     */
    public double getLogFactor() {
        return logFactor;
    }

    /**
     * Obtain the factor multiplying this expression
     */
    public double getFactor() {
        return Math.pow(10, logFactor);
    }

    /**
     * Return a representation of the parsed expression as a list of OneUnits.
     */
    public List<OneUnit>asList() {
        return unitsList;
    }

    /**
     * Returns the unit with the given base-unit.  If there is more
     * than one known abbreviation for a unit (as is the case for
     * "pixel" and "pix"), either may be used, irrespective of which
     * one appeared in the originally parsed string.
     * @param reqUnit the string representation of a base unit we're to find, such as "m"
     * @return the OneUnit representing the presence of this unit in the expression, or null if the unit is not present
     */
    public OneUnit getUnit(String reqUnit) {
        UnitDefinition reqDef = UnitDefinitionMap.getInstance().lookup(reqUnit);

        if (reqDef == null) {
            // not a known unit
            for (OneUnit u : unitsList) {
                if (u.getBaseUnit().equals(reqUnit)) {
                    return u;   // found it -- JUMP OUT
                }
            }
        } else {
            // reqUnit is a known unit: reqDef
            for (OneUnit u : unitsList) {
                UnitDefinition thisDef = u.getBaseUnitDefinition();
                if (thisDef != null && thisDef.equals(reqDef)) {
                    return u;   // found it -- JUMP OUT
                }
            }
        }
        // if we get here, it's because we haven't found this already
        return null;
    }

    /**
     * Indicates whether the parsed expression is composed only of
     * recommended units, in the sense of
     * {@link OneUnit#isRecommendedUnit}.  Recommended units are units which
     * are both recognised and not deprecated.
     * @param syntax the syntax with respect to which the
     * units should be checked
     */
    public boolean allUnitsRecommended(String syntax) {
        for (OneUnit u : asList()) {
            if (! u.isRecommendedUnit(syntax)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Indicates whether the parsed expression is composed only of recognised
     * units, in the sense of {@link OneUnit#isRecognisedUnit}.  That
     * is, an expression composed of only non-deprecated and unknown
     * units has no deprecated units.
     * @param syntax the syntax with respect to which the
     * units should be checked
     */
    public boolean allUnitsRecognised(String syntax) {
        for (OneUnit u : asList()) {
            if (! u.isRecognisedUnit(syntax)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Indicates whether the expression is being used in a way which
     * satisfies any usage constraints,
     * in the sense of {@link OneUnit#satisfiesUsageConstraints}.
     * Principally, this tests
     * whether a unit which may not be used with SI prefixes was
     * provided with a prefix, but there may be other constraints
     * present.
     *
     * <p>An unrecognised unit has no constraints, and so will always
     * satisfy them; this extends to units which are unrecognised in a
     * particular syntax.
     * @param syntax one of the syntaxes of {@link UnitParser}
     */
    public boolean allUsageConstraintsSatisfied(String syntax) {
        for (OneUnit u : asList()) {
            if (! u.satisfiesUsageConstraints(syntax)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Indicates whether the expression is fully conformant with the
     * appropriate recommendations.  That is, all units are
     * recommended, and are being used in a way which is conformant
     * with any usage constraints.
     * @param syntax one of the syntaxes of {@link UnitParser}
     */
    public boolean isFullyConformant(String syntax) {
        for (OneUnit u : asList()) {
            if (! (u.satisfiesUsageConstraints(syntax) && u.isRecommendedUnit(syntax))) {
                return false;
            }
        }
        return true;
    }

    /**
     * Produces a string representation of the unit expression, in a
     * form suitable for display
     */
    public String toString() {
        try {
            return toString(UnitParser.FITS, null);
        } catch (UnitParserException e) {
            throw new AssertionError("UnitExpr.toString(FITS) failed: this shouldn't happen: " + e);
        }
    }

    /**
     * Produces a string representation of the unit expression, in a
     * format appropriate to the given syntax.
     * The syntaxes are those of {@link UnitParser}.  
     * The output is locale-independent (which in practice means it is
     * in the <code>java.text.Locale.US</code> locale).
     * @param syntax the syntax to generate
     */
    public String toString(String syntax)
            throws UnitParserException {
        return toString(syntax, null);
    }

    /**
     * Produces a string representation of the unit expression, in a
     * format appropriate to the given syntax, and with output
     * respecting the conventions of the given locale.
     * The syntaxes are those of {@link UnitParser}.
     * If <code>locale</code> is null, the output is locale-independent
     * (which in practice means it is in the <code>java.text.Locale.US</code> locale);
     * note that this is distinct from the process's default locale.
     * @param syntax the syntax to generate
     * @param locale the locale to use, or null to use a 'no-locale' locale
     */
    public String toString(String syntax, java.util.Locale locale)
            throws UnitParserException {
        UnitWriter w = SyntaxFactory.createWriter(syntax, this);
        if (locale != null) {
            w.setLocale(locale);
        }
        return w.write();
    }

    /**
     * Format the unit expression as a string, in some sort of
     * canonical/unambiguous form.  The format of the output is unspecified: the
     * {@link #toString} method is generally preferred as a
     * way of formatting expressions.
     */
    public String toDebugString() {
        boolean first = true;
        StringBuilder sb = new StringBuilder();
        for (OneUnit u : unitsList) {
            if (first) {
                first = false;
            } else {
                sb.append(' ');
            }
            sb.append(u.toString());
        }
        return sb.toString();
    }
}

        
